from esbonio.sphinx_agent.app import Database


def test_ensure_table_new():
    """Ensure that we can create a table where none existed previously."""

    table = Database.Table(
        "example",
        [
            Database.Column(name="name", dtype="TEXT"),
            Database.Column(name="age", dtype="INTEGER"),
        ],
    )

    database = Database(":memory:")
    database.ensure_table(table)

    cursor = database.db.execute("PRAGMA table_info(example);")
    rows = cursor.fetchall()

    assert rows[0] == (0, "name", "TEXT", 0, None, 0)
    assert rows[1] == (1, "age", "INTEGER", 0, None, 0)


def test_ensure_table_existing():
    """Ensure that ``ensure_table`` is a no-op if the table already exists and is
    compatible with the given definition."""

    table = Database.Table(
        "example",
        [
            Database.Column(name="name", dtype="TEXT"),
            Database.Column(name="age", dtype="INTEGER"),
        ],
    )

    database = Database(":memory:")
    database.ensure_table(table)

    cursor = database.db.execute("PRAGMA schema_version;")
    version = cursor.fetchone()[0]
    assert version == 1

    database.ensure_table(table)

    # Schema version would be incremented if we made changes to the tables
    cursor = database.db.execute("PRAGMA schema_version;")
    version = cursor.fetchone()[0]
    assert version == 1


def test_insert_and_clear_table():
    """Ensure that we can insert values into a table and subsequently clear it."""
    table = Database.Table(
        "example",
        [
            Database.Column(name="name", dtype="TEXT"),
            Database.Column(name="age", dtype="INTEGER"),
        ],
    )

    database = Database(":memory:")
    database.ensure_table(table)
    database.insert_values(table, [("alice", 12), ("bob", 13)])

    cursor = database.db.execute("SELECT * FROM example")
    rows = cursor.fetchall()

    assert rows[0] == ("alice", 12)
    assert rows[1] == ("bob", 13)

    database.clear_table(table)

    cursor = database.db.execute("SELECT * FROM example")
    rows = cursor.fetchall()
    assert len(rows) == 0


def test_ensure_table_update():
    """Ensure that ``ensure_table`` can drop and recreate a table if the schema
    changes.."""

    table = Database.Table(
        "example",
        [
            Database.Column(name="name", dtype="TEXT"),
            Database.Column(name="age", dtype="INTEGER"),
        ],
    )

    database = Database(":memory:")
    database.ensure_table(table)

    cursor = database.db.execute("PRAGMA table_info(example);")
    rows = cursor.fetchall()

    assert rows[0] == (0, "name", "TEXT", 0, None, 0)
    assert rows[1] == (1, "age", "INTEGER", 0, None, 0)

    cursor = database.db.execute("PRAGMA schema_version;")
    version = cursor.fetchone()[0]
    assert version == 1

    table.columns.insert(1, Database.Column(name="address", dtype="TEXT"))
    database.ensure_table(table)

    cursor = database.db.execute("PRAGMA table_info(example);")
    rows = cursor.fetchall()

    assert rows[0] == (0, "name", "TEXT", 0, None, 0)
    assert rows[1] == (1, "address", "TEXT", 0, None, 0)
    assert rows[2] == (2, "age", "INTEGER", 0, None, 0)

    # Schema version would be incremented if we made changes to the tables
    cursor = database.db.execute("PRAGMA schema_version;")
    version = cursor.fetchone()[0]
    assert version == 3



from __future__ import annotations

import inspect
import logging
import sys
import traceback
import typing
from collections.abc import Callable

import sphinx.application
from sphinx import __version__ as __sphinx_version__
from sphinx.util.logging import NAMESPACE as SPHINX_LOG_NAMESPACE

from .. import types
from ..app import Sphinx
from ..config import SphinxConfig
from ..types import Uri
from ..util import send_error
from ..util import send_message

sphinx_logger = logging.getLogger(SPHINX_LOG_NAMESPACE)

# Inject our own 'core' extensions into Sphinx
sphinx.application.builtin_extensions += (
    f"{__name__}.webview",
    f"{__name__}.files",
    f"{__name__}.diagnostics",
    f"{__name__}.symbols",
    f"{__name__}.directives",
    f"{__name__}.roles",
    f"{__name__}.domains",
)


class SphinxHandler:
    """Responsible for implementing the JSON-RPC API exposed by the Sphinx agent."""

    def __init__(self):
        self.app: Sphinx | None = None
        """The sphinx application instance"""

        self._content_overrides: dict[Uri, str] = {}
        """Holds any additional content to inject into a build."""

        self._handlers: dict[str, tuple[type, Callable]] = self._register_handlers()

    def get(self, method: str) -> tuple[type, Callable] | None:
        """Return the handler for the given method - if possible.

        Parameters
        ----------
        method
           The name of the method

        Returns
        -------
        Optional[Tuple[Type, Callable]]
           A tuple where the first element is the type definition
           representing the message body, the second element is the method which
           implements it.

           If ``None``, the given method is unknown.

        """
        return self._handlers.get(method)

    def _register_handlers(self) -> dict[str, tuple[type, Callable]]:
        """Return a map of all the handlers we provide.

        A handler

        - must be a method on this class
        - the must take a single parameter called ``request``
        - the type annotation for that parameter must correspond with a ``XXXRequest``
          class definition from the ``types`` module.

        Returns
        -------
        Dict[str, Tuple[Type, Callable]]
           A dictonary where the keys are the method names implemented by this class.
           Values are a tuple where the first element is the type definition
           representing the message body, the second element is the method which
           implements it.
        """
        handlers: dict[str, tuple[type, Callable]] = {}

        for name in dir(self):
            method_func = getattr(self, name)
            if name.startswith("_") or not inspect.ismethod(method_func):
                continue

            parameters = inspect.signature(method_func).parameters
            if set(parameters.keys()) != {"request"}:
                continue

            request_type = typing.get_type_hints(method_func)["request"]
            if not all(
                [hasattr(request_type, "jsonrpc"), hasattr(request_type, "method")]
            ):
                continue

            handlers[request_type.method] = (request_type, method_func)

        return handlers

    def create_sphinx_app(self, request: types.CreateApplicationRequest):
        """Create a new sphinx application instance."""
        params = request.params

        sphinx_config = SphinxConfig.fromcli(params.command)
        if sphinx_config is None:
            raise ValueError("Invalid build command")

        sphinx_config.config_overrides.update(params.config_overrides)
        sphinx_args = sphinx_config.to_application_args(params.context)
        self.app = Sphinx(**sphinx_args)

        # Connect event handlers.
        # TODO: Sphinx 7.x has introduced a `include-read` event
        # See: https://github.com/sphinx-doc/sphinx/pull/11657
        self.app.connect("env-before-read-docs", self._cb_env_before_read_docs)
        self.app.connect("source-read", self._cb_source_read, priority=0)

        response = types.CreateApplicationResponse(
            id=request.id,
            result=types.SphinxInfo(
                version=__sphinx_version__,
                conf_dir=str(self.app.confdir),
                build_dir=str(self.app.outdir),
                builder_name=self.app.builder.name,
                src_dir=str(self.app.srcdir),
                dbpath=str(self.app.esbonio.db.path),
            ),
            jsonrpc=request.jsonrpc,
        )
        send_message(response)

    def _cb_env_before_read_docs(self, app: Sphinx, env, docnames: list[str]):
        """Used to add additional documents to the "to build" list."""

        is_building = set(docnames)

        for docname in env.found_docs - is_building:
            uri = Uri.for_file(env.doc2path(docname, base=True))
            if uri in self._content_overrides:
                docnames.append(docname)

    def _cb_source_read(self, app: Sphinx, docname: str, source):
        """Called whenever sphinx reads a file from disk."""

        uri = Uri.for_file(app.env.doc2path(docname, base=True))
        if (content := self._content_overrides.get(uri, None)) is not None:
            source[0] = content

    def build_sphinx_app(self, request: types.BuildRequest):
        """Trigger a Sphinx build."""

        if self.app is None:
            send_error(id=request.id, code=-32803, message="Sphinx app not initialized")
            return

        self._content_overrides = {
            Uri.parse(p): content
            for p, content in request.params.content_overrides.items()
        }

        try:
            self.app.build()

            response = types.BuildResponse(
                id=request.id,
                result=types.BuildResult(),
                jsonrpc=request.jsonrpc,
            )
            send_message(response)
        except Exception as exc:
            message = "".join(traceback.format_exception_only(type(exc), exc))
            sphinx_logger.error("sphinx-build failed", exc_info=True)
            send_error(
                id=request.id, code=-32603, message=f"sphinx-build failed: {message}"
            )

        finally:
            self.app._warncount = 0

    def notify_exit(self, request: types.ExitNotification):
        """Sent from the client to signal that the agent should exit."""
        sys.exit(0)



from __future__ import annotations

import enum
from dataclasses import dataclass


@dataclass(frozen=True)
class Position:
    line: int
    character: int


@dataclass(frozen=True)
class Range:
    start: Position
    end: Position


@dataclass(frozen=True)
class Location:
    uri: str
    range: Range


class DiagnosticSeverity(enum.IntEnum):
    Error = 1
    Warning = 2
    Information = 3
    Hint = 4


@dataclass(frozen=True)
class Diagnostic:
    range: Range
    message: str
    severity: DiagnosticSeverity



import argparse
import logging
import sys
import warnings
from collections.abc import Sequence
from logging.handlers import MemoryHandler

from pygls.protocol import default_converter

from .server import EsbonioLanguageServer
from .server import __version__
from .setup import create_language_server


def build_parser() -> argparse.ArgumentParser:
    """Return an argument parser with the default command line options required for
    main.
    """

    cli = argparse.ArgumentParser(description="The Esbonio language server")
    cli.add_argument(
        "-p",
        "--port",
        type=int,
        default=None,
        help="start a TCP instance of the language server listening on the given port.",
    )
    cli.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="print the current version and exit.",
    )

    modules = cli.add_argument_group(
        "modules", "include/exclude language server modules."
    )
    modules.add_argument(
        "-i",
        "--include",
        metavar="MOD",
        action="append",
        default=[],
        dest="included_modules",
        help="include an additional module in the server configuration, can be given multiple times.",
    )
    modules.add_argument(
        "-e",
        "--exclude",
        metavar="MOD",
        action="append",
        default=[],
        dest="excluded_modules",
        help="exclude a module from the server configuration, can be given multiple times.",
    )

    return cli


def main(argv: Sequence[str] | None = None):
    cli = build_parser()
    args = cli.parse_args(argv)

    # Order matters!
    modules = [
        "esbonio.server.features.log",
        "esbonio.server.features.project_manager",
        "esbonio.server.features.sphinx_manager",
        "esbonio.server.features.preview_manager",
        "esbonio.server.features.directives",
        "esbonio.server.features.roles",
        "esbonio.server.features.rst.directives",
        "esbonio.server.features.rst.roles",
        "esbonio.server.features.myst.directives",
        "esbonio.server.features.myst.roles",
        "esbonio.server.features.sphinx_support.diagnostics",
        "esbonio.server.features.sphinx_support.symbols",
        "esbonio.server.features.sphinx_support.directives",
        "esbonio.server.features.sphinx_support.roles",
    ]

    for mod in args.included_modules:
        modules.append(mod)

    for mod in args.excluded_modules:
        if mod in modules:
            modules.remove(mod)

    # Ensure we can capture warnings.
    logging.captureWarnings(True)

    if not sys.warnoptions:
        warnings.simplefilter("default")  # Enable capture of DeprecationWarnings

    # Setup a temporary logging handler that can cache messages until the language server
    # is ready to forward them onto the client.
    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[MemoryHandler(999999, flushLevel=logging.CRITICAL)],
    )

    server = create_language_server(
        EsbonioLanguageServer,
        modules,
        logger=logging.getLogger("esbonio"),
        converter_factory=default_converter,
    )

    if args.port:
        server.start_tcp("localhost", args.port)
    else:
        server.start_io()



from __future__ import annotations

import inspect
import typing

from lsprotocol import types as lsp

from esbonio import server
from esbonio.sphinx_agent import types

from . import providers
from .providers import RoleTargetProvider

if typing.TYPE_CHECKING:
    from collections.abc import Coroutine
    from typing import Any

    from esbonio.server import Uri


class RoleProvider:
    """Base class for role providers."""

    def get_role(
        self, uri: Uri, name: str
    ) -> types.Role | None | Coroutine[Any, Any, types.Role | None]:
        """Return the definition of the given role, if known.

        Parameters
        ----------
        uri
           The uri of the document in which the role name appears

        name
           The name of the role, as the user would type in a document
        """
        return None

    def suggest_roles(
        self, context: server.CompletionContext
    ) -> list[types.Role] | None | Coroutine[Any, Any, list[types.Role] | None]:
        """Givem a completion context, suggest roles that may be used."""
        return None


class RolesFeature(server.LanguageFeature):
    """Backend support for roles.

    It's this language feature's responsibility to provide an API that exposes the
    information a frontend feature may want.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._role_providers: dict[int, RoleProvider] = {}
        self._target_providers: dict[str, RoleTargetProvider] = {}

    def add_role_provider(self, provider: RoleProvider):
        """Register a role provider.

        Parameters
        ----------
        provider
           The role provider
        """
        self._role_providers[id(provider)] = provider

    def add_role_target_provider(self, name: str, provider: RoleTargetProvider):
        """Register a role target provider.

        Parameters
        ----------
        provider
           The role target provider
        """
        if (existing := self._target_providers.get(name)) is not None:
            raise ValueError(
                f"RoleTargetProvider {provider!r} conflicts with existing "
                f"provider: {existing!r}"
            )

        self._target_providers[name] = provider

    async def suggest_roles(
        self, context: server.CompletionContext
    ) -> list[types.Role]:
        """Suggest roles that may be used, given a completion context.

        Parameters
        ----------
        context
           The completion context
        """
        items: list[types.Role] = []

        for provider in self._role_providers.values():
            try:
                result: list[types.Role] | None = None

                aresult = provider.suggest_roles(context)
                if inspect.isawaitable(aresult):
                    result = await aresult

                if result:
                    items.extend(result)
            except Exception:
                name = type(provider).__name__
                self.logger.error("Error in '%s.suggest_roles'", name, exc_info=True)

        return items

    async def get_role(self, uri: Uri, name: str) -> types.Role | None:
        """Return the definition of the given role name.

        Parameters
        ----------
        uri
           The uri of the document in which the role name appears

        name
           The name of the role, as the user would type into a document.

        Returns
        -------
        types.Role | None
           The role's definition, if known
        """
        for provider in self._role_providers.values():
            try:
                result: types.Role | None = None

                aresult = provider.get_role(uri, name)
                if inspect.isawaitable(aresult):
                    result = await aresult

                if result is not None:
                    return result
            except Exception:
                name = type(provider).__name__
                self.logger.error("Error in '%s.get_role'", name, exc_info=True)

        return None

    async def hover_target(
        self, context: server.HoverContext, role_name: str, target: str
    ) -> str | None:
        """Return the hover text for the role's target.

        Parameters
        ----------
        context
           The hover context

        role_name
           The role to get the hover text for

        target
           The role's target

        Returns
        -------
        str | None
           The target's hover, if known
        """
        if (role := await self.get_role(context.uri, role_name)) is None:
            self.logger.debug("Unknown role '%s'", role_name)
            return None

        if not role.target_providers:
            return None

        self.logger.debug(
            "Finding target hover for role: '%s' (%s)",
            role.name,
            role.implementation,
        )

        for spec in role.target_providers:
            if (provider := self._target_providers.get(spec.name)) is None:
                self.logger.error("Unknown target provider: '%s'", spec.name)
                continue

            try:
                result = provider.hover_target(context, target, **spec.kwargs)
                if inspect.isawaitable(result):
                    result = await result

                if result is not None:
                    return result

            except Exception:
                name = type(provider).__name__
                self.logger.error("Error in '%s.hover_target'", name, exc_info=True)

        return None

    async def find_target_definition(
        self, context: server.DefinitionContext, role_name: str, target: str
    ) -> list[lsp.Location] | None:
        """Find the definition of the role's target.

        Parameters
        ----------
        context
           The definition context

        role_name
           The role to find the target definition for

        target
           The role's target

        Returns
        -------
        list[lsp.Location] | None
           The target's definition(s), if known
        """
        if (role := await self.get_role(context.uri, role_name)) is None:
            self.logger.debug("Unknown role '%s'", role_name)
            return None

        if not role.target_providers:
            return None

        self.logger.debug(
            "Finding target definition for role: '%s' (%s)",
            role.name,
            role.implementation,
        )

        for spec in role.target_providers:
            if (provider := self._target_providers.get(spec.name)) is None:
                self.logger.error("Unknown target provider: '%s'", spec.name)
                continue

            try:
                result = provider.find_target_definition(context, target, **spec.kwargs)
                if inspect.isawaitable(result):
                    result = await result

                if result is not None:
                    return result

            except Exception:
                name = type(provider).__name__
                self.logger.error(
                    "Error in '%s.find_target_definition'", name, exc_info=True
                )

        return None

    async def resolve_target_link(
        self, context: server.DocumentLinkContext, role_name: str, target: str
    ) -> None | str | tuple[str, str | None]:
        """Given a role target, resolve the corresponding document uri, if possible.

        Parameters
        ----------
        context
           The document link context

        role_name
           The role to resolve the target for

        target
           The target to resolve
        """
        if (role := await self.get_role(context.uri, role_name)) is None:
            self.logger.debug("Unknown role '%s'", role_name)
            return None

        if not role.target_providers:
            return None

        self.logger.debug(
            "Resolving target link for role: '%s' (%s)",
            role.name,
            role.implementation,
        )

        for spec in role.target_providers:
            if (provider := self._target_providers.get(spec.name)) is None:
                self.logger.error("Unknown target provider: '%s'", spec.name)
                continue

            try:
                result: None | str | tuple[str, str | None] = None

                aresult = provider.resolve_target_link(context, target, **spec.kwargs)
                if inspect.isawaitable(aresult):
                    result = await aresult
                else:
                    result = aresult

                if result is not None:
                    return result

            except Exception:
                name = type(provider).__name__
                self.logger.error(
                    "Error in '%s.resolve_target_link'", name, exc_info=True
                )

        return None

    async def suggest_targets(
        self, context: server.CompletionContext, role_name: str
    ) -> list[lsp.CompletionItem]:
        """Suggest role targets that may be used, given a completion context.

        Parameters
        ----------
        context
           The completion context

        role_name
           The role to suggest targets for
        """
        if (role := await self.get_role(context.uri, role_name)) is None:
            self.logger.debug("Unknown role '%s'", role_name)
            return []

        targets = []
        self.logger.debug(
            "Suggesting targets for role: '%s' (%s)", role.name, role.implementation
        )

        for spec in role.target_providers:
            if (provider := self._target_providers.get(spec.name)) is None:
                self.logger.error("Unknown target provider: '%s'", spec.name)
                continue

            try:
                result: list[lsp.CompletionItem] | None = None

                aresult = provider.suggest_targets(context, **spec.kwargs)
                if inspect.isawaitable(aresult):
                    result = await aresult
                else:
                    result = aresult

                if result is not None:
                    targets.extend(result)

            except Exception:
                name = type(provider).__name__
                self.logger.error("Error in '%s.suggest_targets'", name, exc_info=True)

        return targets


def esbonio_setup(server: server.EsbonioLanguageServer):
    roles = RolesFeature(server)
    roles.add_role_target_provider("filepath", providers.FilepathProvider(server))

    server.add_feature(roles)



from __future__ import annotations

import asyncio
import json

from lsprotocol import types

from esbonio.server import EsbonioLanguageServer
from esbonio.server import LanguageFeature
from esbonio.server import Uri
from esbonio.server.features.project_manager import ProjectManager


class SphinxSymbols(LanguageFeature):
    """Add support for ``textDocument/documentSymbol`` requests"""

    def __init__(self, server: EsbonioLanguageServer, manager: ProjectManager):
        super().__init__(server)
        self.manager = manager

    async def document_symbol(
        self, params: types.DocumentSymbolParams
    ) -> list[types.DocumentSymbol] | None:
        """Called when a document symbols request is received."""

        uri = Uri.parse(params.text_document.uri)
        if (project := self.manager.get_project(uri)) is None:
            return None

        symbols = await project.get_document_symbols(uri)
        if len(symbols) == 0:
            return None

        root: list[types.DocumentSymbol] = []
        index: dict[int, types.DocumentSymbol] = {}

        for id_, name, kind, detail, range_json, parent_id, _ in symbols:
            range_ = self.converter.structure(json.loads(range_json), types.Range)
            symbol = types.DocumentSymbol(
                name=name,
                kind=self.converter.structure(kind, types.SymbolKind),
                range=range_,
                selection_range=range_,
                detail=detail,
            )

            index[id_] = symbol

            if parent_id is None:
                root.append(symbol)
            else:
                parent = index[parent_id]
                if parent.children is None:
                    parent.children = [symbol]
                else:
                    parent.children.append(symbol)  # type: ignore[attr-defined]

        return root

    async def workspace_symbol(
        self, params: types.WorkspaceSymbolParams
    ) -> list[types.WorkspaceSymbol] | None:
        """Called when a workspace symbol request is received."""

        tasks = []
        for project in self.manager.projects.values():
            tasks.append(
                asyncio.create_task(project.get_workspace_symbols(params.query))
            )

        symbols = await asyncio.gather(*tasks)
        result: list[types.WorkspaceSymbol] = []

        for batch in symbols:
            for uri_str, name, kind, detail, range_json, container in batch:
                uri = Uri.parse(uri_str)
                range_ = self.converter.structure(json.loads(range_json), types.Range)

                if detail not in {"", name}:
                    display_name = f"{name} {detail}"
                else:
                    display_name = name

                result.append(
                    types.WorkspaceSymbol(
                        location=types.Location(uri=str(uri), range=range_),
                        name=display_name,
                        kind=self.converter.structure(kind, types.SymbolKind),
                        container_name=container,
                    )
                )

        return result


def esbonio_setup(server: EsbonioLanguageServer, project_manager: ProjectManager):
    symbols = SphinxSymbols(server, project_manager)
    server.add_feature(symbols)



from __future__ import annotations

import importlib
import inspect
import pathlib
import typing
from collections.abc import Iterable

from lsprotocol import types
from pygls.capabilities import get_capability

from . import Uri
from .feature import DocumentLinkContext

if typing.TYPE_CHECKING:
    from .server import EsbonioLanguageServer


def create_language_server(
    server_cls: type[EsbonioLanguageServer], modules: Iterable[str], *args, **kwargs
) -> EsbonioLanguageServer:
    """Create a new language server instance.

    Parameters
    ----------
    server_cls:
       The class definition to create the server from.
    modules:
       The list of modules that should be loaded.
    args, kwargs:
       Any additional arguments that should be passed to the language server's
       constructor.
    """
    server = server_cls(*args, **kwargs)

    for module in modules:
        _load_module(server, module)

    _register_lsp_methods(server)
    _register_completion(server)
    _register_hover(server)
    return server


def _register_lsp_methods(server: EsbonioLanguageServer):
    """Configure method handlers for the portions of the LSP spec we support."""

    @server.feature(types.INITIALIZE)
    async def on_initialize(ls: EsbonioLanguageServer, params: types.InitializeParams):
        text_diagnostic = get_capability(
            params.capabilities, "text_document.diagnostic", None
        )
        workspace_diagnostic = get_capability(
            params.capabilities,
            "workspace.diagnostics",
            types.DiagnosticWorkspaceClientCapabilities(),
        )

        # While we can use the pull-based diagnostic model, we need the client to support
        # `workspace/diagnositc/refresh` requests in order for the experience to be good.
        if text_diagnostic is not None and workspace_diagnostic.refresh_support:
            _register_pull_diagnostics(ls)

        ls.initialize(params)
        await call_features(ls, "initialize", params)

    @server.feature(types.INITIALIZED)
    async def on_initialized(
        ls: EsbonioLanguageServer, params: types.InitializedParams
    ):
        await ls.initialized(params)
        await call_features(ls, "initialized", params)
        ls.ready.set_result(True)

    @server.feature(types.SHUTDOWN)
    async def on_shutdown(ls: EsbonioLanguageServer, params: None):
        ls.lsp_shutdown(params)
        await call_features(ls, "shutdown", params)

    @server.feature(types.TEXT_DOCUMENT_DID_CHANGE)
    async def on_document_change(
        ls: EsbonioLanguageServer, params: types.DidChangeTextDocumentParams
    ):
        await call_features(ls, "document_change", params)

    @server.feature(types.TEXT_DOCUMENT_DID_CLOSE)
    async def on_document_close(
        ls: EsbonioLanguageServer, params: types.DidCloseTextDocumentParams
    ):
        await call_features(ls, "document_close", params)

    @server.feature(types.TEXT_DOCUMENT_DID_OPEN)
    async def on_document_open(
        ls: EsbonioLanguageServer, params: types.DidOpenTextDocumentParams
    ):
        await call_features(ls, "document_open", params)

    @server.feature(types.TEXT_DOCUMENT_DID_SAVE)
    async def on_document_save(
        ls: EsbonioLanguageServer, params: types.DidSaveTextDocumentParams
    ):
        # Record the version number of the document
        doc = ls.workspace.get_text_document(params.text_document.uri)
        doc.saved_version = doc.version or 0  # type: ignore[attr-defined]

        await call_features(ls, "document_save", params)

    @server.feature(types.TEXT_DOCUMENT_DEFINITION)
    async def on_definition(ls: EsbonioLanguageServer, params: types.DefinitionParams):
        uri = params.text_document.uri
        pos = params.position
        doc = ls.workspace.get_text_document(uri)
        language = ls.get_language_at(doc, pos)

        definitions = []

        for cls, feature in ls:
            if not feature.definition_trigger:
                continue

            context = feature.definition_trigger(
                uri=Uri.parse(uri),
                params=params,
                document=doc,
                language=language,
                client_capabilities=ls.client_capabilities,
            )

            if context is None:
                continue

            ls.logger.debug("%s", context)
            name = f"{cls.__name__}"

            try:
                result = feature.definition(context)
                if inspect.isawaitable(result):
                    result = await result
            except Exception:
                ls.logger.exception("Error in '%s.definition' handler", name)
                continue

            if result is not None:
                definitions.extend(result)

        return definitions if len(definitions) > 0 else None

    @server.feature(types.TEXT_DOCUMENT_DOCUMENT_LINK)
    async def on_document_link(
        ls: EsbonioLanguageServer, params: types.DocumentLinkParams
    ):
        uri = params.text_document.uri
        doc = ls.workspace.get_text_document(uri)
        context = DocumentLinkContext(
            uri=Uri.parse(uri), doc=doc, capabilities=ls.client_capabilities
        )
        ls.logger.debug("%s", context)

        links = []
        for cls, feature in ls:
            name = f"{cls.__name__}"

            try:
                result = feature.document_link(context)
                if inspect.isawaitable(result):
                    result = await result

                if isinstance(result, list):
                    links.extend(result)

            except Exception:
                ls.logger.exception("Error in '%s.document_link' handler", name)
                continue

        return links or None

    @server.feature(types.TEXT_DOCUMENT_DOCUMENT_SYMBOL)
    async def on_document_symbol(
        ls: EsbonioLanguageServer, params: types.DocumentSymbolParams
    ):
        result = await return_first_result(ls, "document_symbol", params)
        return result

    @server.feature(types.WORKSPACE_SYMBOL)
    async def on_workspace_symbol(
        ls: EsbonioLanguageServer, params: types.WorkspaceSymbolParams
    ):
        result = await return_first_result(ls, "workspace_symbol", params)
        if len(result) == 0:
            return None

        return result

    @server.feature(types.WORKSPACE_DID_CHANGE_CONFIGURATION)
    async def on_did_change_configuration(
        ls: EsbonioLanguageServer, params: types.DidChangeConfigurationParams
    ):
        ls.logger.debug("%s: %s", types.WORKSPACE_DID_CHANGE_CONFIGURATION, params)
        await ls.configuration.update_workspace_configuration()

    @server.feature(types.WORKSPACE_DID_CHANGE_WATCHED_FILES)
    async def on_did_change_watched_files(
        ls: EsbonioLanguageServer, params: types.DidChangeWatchedFilesParams
    ):
        ls.logger.debug("%s: %s", types.WORKSPACE_DID_CHANGE_WATCHED_FILES, params)
        # TODO: Handle deleted files.
        paths = [pathlib.Path(Uri.parse(event.uri)) for event in params.changes]
        ls.configuration.update_file_configuration(paths)


def _register_completion(server: EsbonioLanguageServer):
    """Configuration completion handlers."""

    trigger_characters: set[str] = set()

    for _, feature in server:
        if feature.completion_trigger is None:
            continue

        trigger_characters.update(feature.completion_trigger.characters)

    @server.feature(
        types.TEXT_DOCUMENT_COMPLETION,
        types.CompletionOptions(
            trigger_characters=list(trigger_characters),
            resolve_provider=True,
        ),
    )
    async def on_completion(ls: EsbonioLanguageServer, params: types.CompletionParams):
        uri = params.text_document.uri
        pos = params.position
        doc = ls.workspace.get_text_document(uri)
        language = ls.get_language_at(doc, pos)

        items = []

        for cls, feature in ls:
            if not feature.completion_trigger:
                continue

            context = feature.completion_trigger(
                uri=Uri.parse(uri),
                params=params,
                document=doc,
                language=language,
                client_capabilities=ls.client_capabilities,
            )

            if context is None:
                continue

            ls.logger.debug("%s", context)
            name = f"{cls.__name__}"

            try:
                result = feature.completion(context)
                if inspect.isawaitable(result):
                    result = await result
            except Exception:
                ls.logger.exception("Error in '%s.complete' handler", name)
                continue

            for item in result or []:
                item.data = {"source_feature": name, **(item.data or {})}  # type: ignore
                items.append(item)

        if len(items) > 0:
            return types.CompletionList(is_incomplete=False, items=items)

    @server.feature(types.COMPLETION_ITEM_RESOLVE)
    def on_completion_resolve(
        ls: EsbonioLanguageServer, item: types.CompletionItem
    ) -> types.CompletionItem:
        # source = (item.data or {}).get("source_feature", "")  # type: ignore
        # feature = ls.get_feature(source)

        # if not feature:
        #     ls.logger.error(
        #         "Unable to resolve completion item, unknown source: '%s'", source
        #     )
        #     return item

        # return feature.completion_resolve(item)
        return item


def _register_pull_diagnostics(server: EsbonioLanguageServer):
    """Configure method handlers for diagnostics."""

    @server.feature(
        types.TEXT_DOCUMENT_DIAGNOSTIC,
        types.DiagnosticOptions(
            identifier="esbonio",
            inter_file_dependencies=True,
            workspace_diagnostics=True,
        ),
    )
    async def on_document_diagnostic(
        ls: EsbonioLanguageServer, params: types.DocumentDiagnosticParams
    ):
        """Handle a ``textDocument/diagnostic`` request."""
        doc_uri = Uri.parse(params.text_document.uri).resolve()
        items = []

        for (_, uri), diags in ls._diagnostics.items():
            if uri.resolve() == doc_uri:
                items.extend(diags)

        # TODO: Detect no changes and send 'unchanged' responses
        return types.RelatedFullDocumentDiagnosticReport(
            items=items,
            kind=types.DocumentDiagnosticReportKind.Full,
        )

    @server.feature(types.WORKSPACE_DIAGNOSTIC)
    async def on_workspace_diagnostic(
        ls: EsbonioLanguageServer, params: types.WorkspaceDiagnosticParams
    ):
        """Handle a ``workspace/diagnostic`` request."""
        diagnostics: dict[Uri, list[types.Diagnostic]] = {}

        for (_, uri), diags in ls._diagnostics.items():
            diagnostics.setdefault(uri, []).extend(diags)

        # TODO: Detect no changes and send 'unchanged' responses
        reports = []
        for uri, items in diagnostics.items():
            reports.append(
                types.WorkspaceFullDocumentDiagnosticReport(
                    uri=str(uri),
                    items=items,
                    kind=types.DocumentDiagnosticReportKind.Full,
                )
            )

        return types.WorkspaceDiagnosticReport(items=reports)


def _register_hover(server: EsbonioLanguageServer):
    """Configure the handlers for documentation hovers."""

    @server.feature(types.TEXT_DOCUMENT_HOVER)
    async def on_hover(ls: EsbonioLanguageServer, params: types.HoverParams):
        uri = params.text_document.uri
        pos = params.position
        doc = ls.workspace.get_text_document(uri)
        language = ls.get_language_at(doc, pos)

        for cls, feature in ls:
            if not feature.hover_trigger:
                continue

            context = feature.hover_trigger(
                uri=Uri.parse(uri),
                params=params,
                document=doc,
                language=language,
                client_capabilities=ls.client_capabilities,
            )

            if context is None:
                continue

            ls.logger.debug("%s", context)
            name = f"{cls.__name__}"

            try:
                result = feature.hover(context)
                if inspect.isawaitable(result):
                    result = await result
            except Exception:
                ls.logger.exception("Error in '%s.hover' handler", name)
                continue

            if result is not None:
                return result

        return None


async def call_features(ls: EsbonioLanguageServer, method: str, *args, **kwargs):
    """Call all features."""

    for cls, feature in ls:
        try:
            impl = getattr(feature, method)

            result = impl(*args, **kwargs)
            if inspect.isawaitable(result):
                await result

        except Exception:
            name = f"{cls.__name__}"
            ls.logger.error("Error in '%s.%s' handler", name, method, exc_info=True)


async def return_first_result(ls: EsbonioLanguageServer, method: str, *args, **kwargs):
    """Call all features, returning the first non ``None`` result we find."""

    for cls, feature in ls:
        try:
            impl = getattr(feature, method)

            result = impl(*args, **kwargs)
            if inspect.isawaitable(result):
                result = await result

            if result is not None:
                return result

        except Exception:
            name = f"{cls.__name__}"
            ls.logger.error("Error in '%s.%s' handler", name, method, exc_info=True)


def _load_module(server: EsbonioLanguageServer, modname: str):
    """Load an extension module by calling its ``esbonio_setup`` function, if it exists."""

    try:
        module = importlib.import_module(modname)
    except ImportError:
        server.logger.error("Unable to import module '%s'", modname, exc_info=True)
        return None

    setup = getattr(module, "esbonio_setup", None)
    if setup is None:
        server.logger.debug(
            "Skipping module '%s', missing 'esbonio_setup' function", modname
        )
        return None

    server.load_extension(modname, setup)



from __future__ import annotations

import typing

import pytest
from lsprotocol import types
from pygls.workspace import TextDocument
from pytest_lsp import client_capabilities

from esbonio import server
from esbonio.server.features.roles import completion
from esbonio.server.testing import range_from_str
from esbonio.sphinx_agent.types import RST_ROLE
from esbonio.sphinx_agent.types import Role

if typing.TYPE_CHECKING:
    from typing import Literal


VSCODE = "visual-studio-code"
NVIM = "neovim"
PATTERNS = {"rst": RST_ROLE}


@pytest.mark.parametrize(
    "client,language,insert_behavior,role,text,character,expected",
    [
        (
            VSCODE,
            "rst",
            "replace",
            Role("ref", "sphinx.roles.XRefRole"),
            ":",
            None,
            types.CompletionItem(
                label="ref",
                detail="sphinx.roles.XRefRole",
                kind=types.CompletionItemKind.Function,
                filter_text=":ref:",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:1"), new_text=":ref:"
                ),
                data={"completion_type": "role"},
            ),
        ),
        (
            VSCODE,
            "rst",
            "replace",
            Role("ref", "sphinx.roles.XRefRole"),
            ":r",
            None,
            types.CompletionItem(
                label="ref",
                detail="sphinx.roles.XRefRole",
                kind=types.CompletionItemKind.Function,
                filter_text=":ref:",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:2"), new_text=":ref:"
                ),
                data={"completion_type": "role"},
            ),
        ),
        (
            VSCODE,
            "rst",
            "replace",
            Role("ref", "sphinx.roles.XRefRole"),
            ":doc",
            None,
            types.CompletionItem(
                label="ref",
                detail="sphinx.roles.XRefRole",
                kind=types.CompletionItemKind.Function,
                filter_text=":ref:",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:4"), new_text=":ref:"
                ),
                data={"completion_type": "role"},
            ),
        ),
        (
            VSCODE,
            "rst",
            "replace",
            Role("cpp:func", "sphinx.domains.cpp.CPPXRefRole"),
            ":c",
            None,
            types.CompletionItem(
                label="cpp:func",
                detail="sphinx.domains.cpp.CPPXRefRole",
                kind=types.CompletionItemKind.Function,
                filter_text=":cpp:func:",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:2"), new_text=":cpp:func:"
                ),
                data={"completion_type": "role"},
            ),
        ),
        (
            VSCODE,
            "rst",
            "replace",
            Role("cpp:func", "sphinx.domains.cpp.CPPXRefRole"),
            ":cpp:f",
            None,
            types.CompletionItem(
                label="cpp:func",
                detail="sphinx.domains.cpp.CPPXRefRole",
                kind=types.CompletionItemKind.Function,
                filter_text=":cpp:func:",
                insert_text_format=types.InsertTextFormat.PlainText,
                text_edit=types.TextEdit(
                    range=range_from_str("0:0-0:6"), new_text=":cpp:func:"
                ),
                data={"completion_type": "role"},
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Role("ref", "sphinx.roles.XRefRole"),
            ":",
            None,
            types.CompletionItem(
                label="ref",
                detail="sphinx.roles.XRefRole",
                kind=types.CompletionItemKind.Function,
                insert_text="ref:",
                insert_text_format=types.InsertTextFormat.PlainText,
                data={"completion_type": "role"},
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Role("ref", "sphinx.roles.XRefRole"),
            ":r",
            None,
            types.CompletionItem(
                label="ref",
                detail="sphinx.roles.XRefRole",
                kind=types.CompletionItemKind.Function,
                insert_text="ref:",
                insert_text_format=types.InsertTextFormat.PlainText,
                data={"completion_type": "role"},
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Role("ref", "sphinx.roles.XRefRole"),
            ":doc",
            None,
            None,
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Role("cpp:func", "sphinx.domains.cpp.CPPXRefRole"),
            ":c",
            None,
            types.CompletionItem(
                label="cpp:func",
                detail="sphinx.domains.cpp.CPPXRefRole",
                kind=types.CompletionItemKind.Function,
                insert_text="cpp:func:",
                insert_text_format=types.InsertTextFormat.PlainText,
                data={"completion_type": "role"},
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Role("cpp:func", "sphinx.domains.cpp.CPPXRefRole"),
            ":cpp:",
            None,
            types.CompletionItem(
                label="cpp:func",
                detail="sphinx.domains.cpp.CPPXRefRole",
                kind=types.CompletionItemKind.Function,
                insert_text="func:",
                insert_text_format=types.InsertTextFormat.PlainText,
                data={"completion_type": "role"},
            ),
        ),
        (
            VSCODE,
            "rst",
            "insert",
            Role("cpp:func", "sphinx.domains.cpp.CPPXRefRole"),
            ":cpp:f",
            None,
            types.CompletionItem(
                label="cpp:func",
                detail="sphinx.domains.cpp.CPPXRefRole",
                kind=types.CompletionItemKind.Function,
                insert_text="func:",
                insert_text_format=types.InsertTextFormat.PlainText,
                data={"completion_type": "role"},
            ),
        ),
    ],
)
def test_render_role_completion(
    client: str,
    language: Literal["rst", "markdown"],
    insert_behavior: Literal["insert", "replace"],
    role: Role,
    text: str,
    character: int | None,
    expected: types.CompletionItem | None,
):
    """Ensure that we can render role completions correctly.

    Parameters
    ----------
    client
       The name of the client to use.

       This will be passed to the ``client_capabilities`` function from ``pytest_lsp``
       and controls which capabilities (e.g. snippet support) will be available.

    language
       The language in which the completion item will be inserted.

    insert_behavior
       How the completion item should behave when inserted

    role
       The role to generate the completion item for

    text
       The text used to help generate the completion context.

    character
       The character column at which the request is being made.
       If ``None``, it will be assumed that the request is being made at the end of ``text``

    expected
       The expected result.
    """
    match = PATTERNS[language].match(text)
    if not match:
        raise ValueError(f"'{text}' is not valid in this context")

    line = 0
    character = len(text) if character is None else character
    uri = "file:///test.txt"

    context = server.CompletionContext(
        uri=server.Uri.parse(uri),
        doc=TextDocument(uri=uri),
        match=match,
        position=types.Position(line=line, character=character),
        language=language,
        capabilities=client_capabilities(client),
    )

    render_func = completion.get_role_renderer(language, insert_behavior)
    assert render_func is not None

    item = render_func(context, role)
    if expected is None:
        assert item is None
    else:
        assert item == expected



from __future__ import annotations

from typing import Literal

import attrs

SyncScrollOptions = Literal[
    "bothWays", "editorWithPreview", "previewWithEditor", "disabled"
]


@attrs.define
class PreviewConfig:
    """Configuration settings for previews."""

    bind: str = attrs.field(default="localhost")
    """The network interface to bind to, defaults to ``localhost``"""

    http_port: int = attrs.field(default=0)
    """The port to host the HTTP server on. If ``0`` a random port number will be
    chosen"""

    ws_port: int = attrs.field(default=0)
    """The port to host the WebSocket server on. If ``0`` a random port number will be
    chosen"""

    synchronize_scroll: SyncScrollOptions = attrs.field(default="bothWays")
    """Control how synchronised scrolling works"""

    show_line_markers: bool = attrs.field(default=False)
    """If set, render the source line markers in the preview"""



from __future__ import annotations

from esbonio.server import EsbonioLanguageServer
from esbonio.server.features.project_manager import ProjectManager

from .client import ClientState
from .client import SphinxClient
from .client_subprocess import make_subprocess_sphinx_client
from .config import SphinxConfig
from .config import SubProcess
from .manager import RestartSphinxParams
from .manager import SphinxManager

__all__ = [
    "ClientState",
    "SphinxClient",
    "SphinxConfig",
    "SphinxManager",
    "SubProcess",
]


def esbonio_setup(server: EsbonioLanguageServer, project_manager: ProjectManager):
    manager = SphinxManager(make_subprocess_sphinx_client, project_manager, server)
    server.add_feature(manager)

    @server.command("esbonio.sphinx.restart")
    async def restart_client(
        ls: EsbonioLanguageServer, params: RestartSphinxParams, *args
    ):
        ls.logger.debug("esbonio.sphinx.restart: %s", params)
        await manager.restart_client(params.id)



from __future__ import annotations

import typing

from lsprotocol import types as lsp

from esbonio import server
from esbonio.server.features import directives
from esbonio.server.features.project_manager import ProjectManager
from esbonio.sphinx_agent import types

if typing.TYPE_CHECKING:
    from esbonio.server import Uri
    from esbonio.server.features.project_manager import Project


class SphinxDirectives(directives.DirectiveProvider):
    """Support for directives in a sphinx project."""

    def __init__(self, manager: ProjectManager):
        self.manager = manager

    async def get_default_domain(self, project: Project, uri: Uri) -> str:
        """Get the name of the default domain for the given document"""

        # Does the document have a default domain set?
        results = await project.find_symbols(
            uri=str(uri.resolve()),
            kind=lsp.SymbolKind.Class.value,
            detail="default-domain",
        )
        if len(results) > 0:
            default_domain = results[0][1]
        else:
            default_domain = None

        primary_domain = await project.get_config_value("primary_domain")
        return default_domain or primary_domain or "py"

    async def get_directive(self, uri: Uri, name: str) -> types.Directive | None:
        """Return the directive with the given name."""

        if (project := self.manager.get_project(uri)) is None:
            return None

        if (directive := await project.get_directive(name)) is not None:
            return directive

        if (directive := await project.get_directive(f"std:{name}")) is not None:
            return directive

        default_domain = await self.get_default_domain(project, uri)
        return await project.get_directive(f"{default_domain}:{name}")

    async def suggest_directives(
        self, context: server.CompletionContext
    ) -> list[types.Directive] | None:
        """Given a completion context, suggest directives that may be used."""

        if (project := self.manager.get_project(context.uri)) is None:
            return None

        # Does the document have a default domain set?
        results = await project.find_symbols(
            uri=str(context.uri.resolve()),
            kind=lsp.SymbolKind.Class.value,
            detail="default-domain",
        )
        if len(results) > 0:
            default_domain = results[0][1]
        else:
            default_domain = None

        primary_domain = await project.get_config_value("primary_domain")
        active_domain = default_domain or primary_domain or "py"

        result: list[types.Directive] = []
        for name, implementation in await project.get_directives():
            # std: directives can be used unqualified
            if name.startswith("std:"):
                short_name = name.replace("std:", "")
                result.append(
                    types.Directive(name=short_name, implementation=implementation)
                )

            # Also suggest unqualified versions of directives from the currently active domain.
            elif name.startswith(f"{active_domain}:"):
                short_name = name.replace(f"{active_domain}:", "")
                result.append(
                    types.Directive(name=short_name, implementation=implementation)
                )

            result.append(types.Directive(name=name, implementation=implementation))

        return result


def esbonio_setup(
    project_manager: ProjectManager,
    directive_feature: directives.DirectiveFeature,
):
    provider = SphinxDirectives(project_manager)
    directive_feature.add_directive_provider(provider)



import logging
import typing
from typing import IO

from docutils import nodes
from docutils.core import Publisher
from docutils.io import NullOutput
from docutils.io import StringInput
from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives
from docutils.readers.standalone import Reader
from docutils.utils import Reporter
from sphinx.config import Config
from sphinx.io import SphinxDummyWriter
from sphinx.util import get_filetype
from sphinx.util.docutils import CustomReSTDispatcher

from .. import types
from ..app import Database
from ..app import Sphinx
from ..util import as_json
from . import sphinx_logger

SYMBOLS_TABLE = Database.Table(
    "symbols",
    [
        Database.Column(name="uri", dtype="TEXT"),  # TODO: Replace with foreign key??
        Database.Column(name="id", dtype="INTEGER"),
        Database.Column(name="name", dtype="TEXT"),
        Database.Column(name="kind", dtype="INTEGER"),
        Database.Column(name="detail", dtype="TEXT"),
        Database.Column(name="range", dtype="JSON"),
        Database.Column(name="parent_id", dtype="INTEGER"),
        Database.Column(name="order_id", dtype="INTEGER"),
    ],
)

# SymbolKinds see: https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#symbolKind
ClassSymbol = 5
StringSymbol = 15


def init_db(app: Sphinx, config: Config):
    app.esbonio.db.ensure_table(SYMBOLS_TABLE)


def update_symbols(app: Sphinx, docname: str, source):
    """Update the symbols defined in the given file."""

    filename = app.env.doc2path(docname)
    filetype = get_filetype(app.config.source_suffix, filename)

    reader = LoggingDoctreeReader(sphinx_logger)
    parser = app.registry.create_source_parser(app, filetype)

    # Reuse the settings from Sphinx's publisher.
    sphinx_pub = app.registry.get_publisher(app, filetype)
    settings = sphinx_pub.settings

    with disable_roles_and_directives():
        publisher = Publisher(
            parser=parser,
            reader=reader,
            writer=SphinxDummyWriter(),
            source_class=StringInput,
            destination=NullOutput(),
        )
        publisher.settings = settings
        publisher.set_source(source="\n".join(source), source_path=str(filename))
        publisher.publish()
        document = publisher.document

    visitor = SymbolVisitor(document)
    document.walkabout(visitor)  # type: ignore[union-attr]

    uri = str(types.Uri.for_file(app.env.doc2path(docname, base=True)).resolve())
    symbols = [(uri, *s) for s in visitor.symbols]

    app.esbonio.db.clear_table(SYMBOLS_TABLE, uri=uri)
    app.esbonio.db.insert_values(SYMBOLS_TABLE, symbols)


def setup(app: Sphinx):
    app.connect("config-inited", init_db)

    # We want this to happen very early so that we see the symbols as written in the
    # file - before any fancy extensions make their changes.
    #
    # The only handler with a higher priority (i.e. 0), should be the handler we use
    # to override the contents of the file so that we stay in sync with the language
    # client.
    app.connect("source-read", update_symbols, priority=1)

    # TODO: Sphinx 7.x+ support
    # app.connect("include-read")


class a_directive(nodes.Element, nodes.Inline):
    """Represents a directive."""


class a_role(nodes.Element):
    """Represents a role."""

    def astext(self):
        return self["text"]


def dummy_role(name, rawtext, text, lineno, inliner, options=None, content=None):
    node = a_role()
    node.line = lineno

    match = types.RST_ROLE.match(rawtext)
    if match is None:
        node.attributes["text"] = rawtext
    else:
        node.attributes.update(match.groupdict())
        node.attributes["text"] = match.group(0)

    return [node], []


def run_dummy_directive(self):
    """This is a dummy implementation"""
    node = a_directive()
    node.line = self.lineno

    node.attributes["name"] = self.name
    node.attributes["options"] = self.options

    if len(self.arguments) > 0:
        node.attributes["argument"] = self.arguments[0]
    else:
        node.attributes["argument"] = None

    if self.content:
        # This is essentially what `nested_parse_with_titles` does in Sphinx.
        # But by passing the content_offset to state.nested_parse we ensure any line
        # numbers remain relative to the start of the current file.
        current_titles = self.state.memo.title_styles
        current_sections = self.state.memo.section_level
        self.state.memo.title_styles = []
        self.state.memo.section_level = 0
        try:
            self.state.nested_parse(
                self.content, self.content_offset, node, match_titles=1
            )
        finally:
            self.state.memo.title_styles = current_titles
            self.state.memo.section_level = current_sections

    return [node]


class disable_roles_and_directives(CustomReSTDispatcher):
    """Disables all roles and directives from being expanded.

    The ``CustomReSTDispactcher`` from Sphinx is *very* cool.  It provides a way to
    override the mechanism used to lookup roles and directives during parsing!

    It's perfect for temporarily replacing all role and directive implementations with
    dummy ones. Parsing an rst document with these dummy implementations gives us
    something equivalent to an abstract syntax tree.

    """

    def directive(self, directive_name, language_module, document):
        # We still have access to the original dispatch mechanism
        # This allows us to adapt our dummy directive to match the "shape" of the real
        # directive implementation!
        impl, _ = self.directive_func(directive_name, language_module, document)
        if impl is None:
            # Fallback to some sensible defaults.
            has_content = True
            option_spec = None
            required_arguments = 0
            optional_arguments = 1
            final_argument_whitespace = True
        else:
            # Mimic the "shape" of the real directive
            if impl.option_spec is None:
                option_spec = None
            else:
                option_spec = {o: directives.unchanged for o in impl.option_spec}

            # It probably doesn't make sense to copy these values, as often the user's
            # usage of a directive will be incorrect.
            required_arguments = 0  # impl.required_arguments
            has_content = True  # impl.has_content

            optional_arguments = 1
            final_argument_whitespace = True

        attrs = {
            "has_content": has_content,
            "option_spec": option_spec,
            "required_arguments": required_arguments,
            "optional_arguments": optional_arguments,
            "final_argument_whitespace": final_argument_whitespace,
            "run": run_dummy_directive,
        }
        return type("DummyDirective", (Directive,), attrs), []

    def role(self, role_name, language_module, lineno, reporter):
        return dummy_role, []


class SymbolVisitor(nodes.NodeVisitor):
    """Used to extract all the symbols from a document."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.symbols: list[types.Symbol] = []
        """Holds the symbols for the document"""

        self.parents: list[int] = []
        """Holds the ids of the current hierarchy."""

        self.order: list[int] = [0]
        """Holds the current position at each level of the hierarchy."""

    def push_symbol(self, name: str, kind: int, range_: types.Range, detail: str = ""):
        """Push another symbol onto the hierarchy.

        Parameters
        ----------
        name
           The name of the symbol

        kind
           The kind of symbol

        range_
           The range denoting where the symbol is in the document.

        detail
           Additional details about the symbol
        """
        symbol_id = len(self.symbols)
        parent_id = self.parents[-1] if len(self.parents) > 0 else None

        order_id = self.order[-1]
        self.order[-1] += 1

        self.parents.append(symbol_id)
        self.order.append(0)

        self.symbols.append(
            (symbol_id, name, kind, detail, as_json(range_), parent_id, order_id)
        )

    def pop_symbol(self):
        """Pop the current level of the hierarchy."""
        self.parents.pop()
        self.order.pop()

    def visit_section(self, node: nodes.Node) -> None:
        name = node.children[0].astext()
        line = (node.line or 1) - 1
        range_ = types.Range(
            start=types.Position(line=line, character=0),
            end=types.Position(line=line, character=len(name) - 1),
        )

        self.push_symbol(name, StringSymbol, range_)

    def depart_section(self, node: nodes.Node) -> None:
        self.pop_symbol()

    def visit_a_directive(self, node: a_directive):
        argument = node.attributes.get("argument", None)
        directive = node.attributes.get("name", "<<unknown>>")

        name = None
        detail = ""

        if argument is not None and len(argument) > 0:
            name = argument

        if directive is not None and len(directive) > 0:
            detail = directive

        if name is None:
            name = directive

        line = (node.line or 1) - 1
        range_ = types.Range(
            start=types.Position(line=line, character=0),
            end=types.Position(line=line, character=len(name) - 1),
        )

        self.push_symbol(name, ClassSymbol, range_, detail=detail)

    def depart_a_directive(self, node: nodes.Node):
        self.pop_symbol()

    # TODO: Enable symbols for roles
    #       However the reported line numbers can be inaccurate...
    def visit_a_role(self, node: nodes.Node) -> None: ...

    def depart_a_role(self, node: nodes.Node) -> None: ...

    # TODO: Enable symbols for definition list items
    #       However the reported line numbers appear to be inaccurate...

    def visit_Text(self, node: nodes.Node) -> None:
        pass

    def depart_Text(self, node: nodes.Node) -> None:
        pass

    def unknown_visit(self, node: nodes.Node) -> None:
        pass

    def unknown_departure(self, node: nodes.Node) -> None:
        pass


class LogStream:
    def __init__(self, logger: logging.Logger):
        self.logger = logger

    def write(self, text: str):
        self.logger.debug(text)


class LogReporter(Reporter):
    """A docutils reporter that writes to the given logger."""

    def __init__(
        self,
        logger: logging.Logger,
        source: str,
        report_level: int,
        halt_level: int,
        debug: bool,
        error_handler: str,
    ) -> None:
        stream = typing.cast(IO, LogStream(logger))
        super().__init__(
            source, report_level, halt_level, stream, debug, error_handler=error_handler
        )  # type: ignore


class LoggingDoctreeReader(Reader):
    """A reader that replaces the default reporter with one that redirects."""

    def __init__(self, logger: logging.Logger, *args, **kwargs):
        self.logger = logger
        super().__init__(*args, **kwargs)

    def new_document(self) -> nodes.document:
        document = super().new_document()

        reporter = document.reporter
        document.reporter = LogReporter(
            self.logger,
            reporter.source,
            reporter.report_level,
            reporter.halt_level,
            reporter.debug_flag,
            reporter.error_handler,
        )

        return document



import pathlib

import pytest

from esbonio.server import Uri

TEST_DIR = pathlib.Path(__file__).parent


def pytest_addoption(parser):
    """Add additional cli arguments to pytest."""

    group = parser.getgroup("esbonio")
    group.addoption(
        "--enable-devtools",
        dest="enable_devtools",
        action="store_true",
        help="enable lsp-devtools integrations",
    )


@pytest.hookimpl(tryfirst=True)
def pytest_report_header(config: pytest.Config):
    """Report additional information in pytest's output header"""
    lines = []

    try:
        from sphinx import __version__

        lines.append(f"sphinx: v{__version__}")
    except ImportError:
        lines.append("sphinx: none")

    return lines


@pytest.fixture(scope="session")
def uri_for():
    """Helper function for returning the uri for a given file in the ``tests/``
    directory."""

    def fn(*args):
        path = (TEST_DIR / pathlib.Path(*args)).resolve()
        assert path.exists(), f"{path} does not exist"
        return Uri.for_file(str(path))

    return fn



from __future__ import annotations

import asyncio
import traceback
import typing
import uuid
from functools import partial

import attrs
import lsprotocol.types as lsp

from esbonio import server
from esbonio.server import Uri
from esbonio.sphinx_agent import types

from .client import ClientState
from .config import SphinxConfig
from .config import SubProcess

if typing.TYPE_CHECKING:
    from collections.abc import Callable

    from esbonio.server.features.project_manager import ProjectManager

    from .client import SphinxClient

    SphinxClientFactory = Callable[["SphinxManager", "SphinxConfig"], "SphinxClient"]


@attrs.define
class ClientCreatedNotification:
    """The payload of a ``sphinx/clientCreated`` notification"""

    id: str
    """The client's id"""

    scope: str
    """The scope at which the client was created."""

    config: SphinxConfig
    """The final configuration."""


@attrs.define
class AppCreatedNotification:
    """The payload of a ``sphinx/appCreated`` notification"""

    id: str
    """The client's id"""

    application: types.SphinxInfo
    """Details about the created application."""


@attrs.define
class ClientErroredNotification:
    """The payload of a ``sphinx/clientErrored`` notification"""

    id: str
    """The client's id"""

    error: str
    """Short description of the error."""

    detail: str
    """Detailed description of the error."""


@attrs.define
class ClientDestroyedNotification:
    """The payload of ``sphinx/clientDestroyed`` notification."""

    id: str
    """The client's id"""


@attrs.define
class SphinxBuildTriggers:
    """Valid configuration options for Sphinx build triggers."""

    on_save: bool = attrs.field(default=True)
    """Trigger a build when a file is saved."""

    on_change: bool | float = attrs.field(default=2.0)
    """Trigger a build each time a file has changed, with a configurable delay."""


@attrs.define
class ManagerConfig:
    """Configuration options for the sphinx manager."""

    build_triggers: SphinxBuildTriggers = attrs.field(factory=SphinxBuildTriggers)
    """Options controlling when to trigger a Sphinx build."""


@attrs.define
class RestartSphinxParams:
    """Parameters for the ``esbonio.sphinx.restart`` command"""

    id: str
    """The id of the sphinx client to restart"""


class SphinxManager(server.LanguageFeature):
    """Responsible for managing Sphinx application instances."""

    def __init__(
        self,
        client_factory: SphinxClientFactory,
        project_manager: ProjectManager,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.client_factory = client_factory
        """Used to create new Sphinx client instances."""

        self.project_manager = project_manager
        """The project manager instance to use."""

        self.clients: dict[str, SphinxClient | None] = {
            # Prevent any clients from being created in the global scope.
            "": None,
        }
        """Holds currently active Sphinx clients."""

        self.config: ManagerConfig = ManagerConfig()
        """The SphinxManager's configuration."""

        self._events = server.EventSource(self.logger)
        """The SphinxManager can emit events."""

        self._pending_builds: dict[str, asyncio.Task] = {}
        """Holds tasks that will trigger a build after a given delay if not cancelled."""

        self._progress_tokens: dict[str, str] = {}
        """Holds work done progress tokens."""

    def add_listener(self, event: str, handler):
        self._events.add_listener(event, handler)

    def initialize(self, params: lsp.InitializeParams):
        """Called once the initial handshake between client and server has finished."""

        self.converter.register_structure_hook(bool | float, lambda obj, _: obj)
        self.converter.register_structure_hook(
            SubProcess, _structure_list_or_subprocess
        )

        self.configuration.subscribe(
            "esbonio.sphinx", ManagerConfig, self.update_configuration
        )

    def update_configuration(self, event: server.ConfigChangeEvent[ManagerConfig]):
        """Called when the user's configuration is updated."""
        self.config = event.value

    async def document_change(self, params: lsp.DidChangeTextDocumentParams):
        if (uri := Uri.parse(params.text_document.uri)) is None:
            return

        client = await self.get_client(uri)
        if client is None:
            return

        if (delay := self.config.build_triggers.on_change) is False:
            return

        # Cancel any existing pending builds
        if (task := self._pending_builds.pop(client.id, None)) is not None:
            task.cancel()

        self._pending_builds[client.id] = asyncio.create_task(
            self.trigger_build_after(
                uri,
                client.id,
                delay=max(float(delay), 1.0),  # Enforce a minimum 1s delay
            )
        )

    async def document_open(self, params: lsp.DidOpenTextDocumentParams):
        # Ensure that a Sphinx app instance is created the first time a document in a
        # given project is opened.
        if (uri := Uri.parse(params.text_document.uri)) is not None:
            await self.get_client(uri)

    async def document_save(self, params: lsp.DidSaveTextDocumentParams):
        if (uri := Uri.parse(params.text_document.uri)) is None:
            return

        client = await self.get_client(uri)
        if client is None:
            return

        if not self.config.build_triggers.on_save:
            return

        # Cancel any existing pending builds
        if (task := self._pending_builds.pop(client.id, None)) is not None:
            task.cancel()

        self.logger.debug("Build triggered on save")
        await self.trigger_build(uri)

    async def shutdown(self, params: None):
        """Called when the server is instructed to ``shutdown``."""

        # Stop any existing clients.
        tasks = []
        for client in self.clients.values():
            if client:
                self.logger.debug("Stopping SphinxClient: %s", client)
                tasks.append(asyncio.create_task(client.stop()))

        await asyncio.gather(*tasks)

    async def trigger_build_after(self, uri: Uri, app_id: str, delay: float):
        """Trigger a build for the given uri after the given delay."""
        await asyncio.sleep(delay)

        self._pending_builds.pop(app_id)
        self.logger.debug("Build triggered after %ss delay", delay)
        await self.trigger_build(uri)

    async def trigger_build(self, uri: Uri):
        """Trigger a build for the relevant Sphinx application for the given uri."""

        client = await self.get_client(uri)
        if client is None:
            return

        if client.state not in {ClientState.Running}:
            self.logger.debug("Skipping build, client is %s", client.state)
            return

        if (project := self.project_manager.get_project(uri)) is None:
            self.logger.debug("Skipping build, project is None")
            return

        # Pass through any unsaved content to the Sphinx agent.
        content_overrides: dict[str, str] = {}
        known_src_uris = await project.get_src_uris()

        for src_uri in known_src_uris:
            doc = self.server.workspace.get_text_document(str(src_uri))
            doc_version = doc.version or 0
            saved_version = getattr(doc, "saved_version", 0)

            if saved_version < doc_version:
                content_overrides[str(src_uri)] = doc.source

        try:
            result = await client.build(content_overrides=content_overrides)

            # Notify listeners.
            self._events.trigger("build", client, result)

        except Exception as exc:
            self.server.window_show_message(
                lsp.ShowMessageParams(message=f"{exc}", type=lsp.MessageType.Error)
            )
            return

    async def restart_client(self, client_id: str):
        """Restart the client with the given id"""
        for client in self.clients.values():
            if client is None:
                continue

            if client.id != client_id:
                continue

            try:
                await client.restart()
            except Exception:
                self.logger.exception("Unable to restart sphinx client")

            break

        else:
            self.logger.error(f"No client with id {client_id!r} available to restart")

    async def get_client(self, uri: Uri) -> SphinxClient | None:
        """Given a uri, return the relevant sphinx client instance for it."""

        scope = self.server.configuration.scope_for(uri)
        if scope not in self.clients:
            self.logger.debug("No client found, creating new subscription")
            self.server.configuration.subscribe(
                "esbonio.sphinx",
                SphinxConfig,
                partial(self._create_or_replace_client, uri),
                scope=uri,
            )

            # It's possible for this code path to be hit multiple times in quick
            # succession e.g. on a fresh server start with N .rst files already open,
            # creating the opportunity to accidentally spawn N duplicated clients!
            #
            # To prevent this, store a `None` at this scope, all being well it will be
            # replaced with the actual client instance when the
            # `_create_or_replace_client` callback runs.
            self.clients[scope] = None

            # The first few callers in a given scope will miss out, but that shouldn't
            # matter too much
            return None

        if (client := self.clients[scope]) is None:
            self.logger.debug("No applicable client for uri: %s", uri)
            return None

        return await client

    async def _create_or_replace_client(
        self, uri: Uri, event: server.ConfigChangeEvent[SphinxConfig]
    ):
        """Create or replace thesphinx client instance for the given config.

        Parameters
        ----------
        uri
           The uri for which the sphinx client was originally created for

        event
           The configuration change event
        """

        config = event.value

        # Do not try and create clients in the global scope
        if event.scope == "":
            return

        # If there was a previous client, stop it.
        if (previous_client := self.clients.pop(event.scope, None)) is not None:
            self.server.protocol.notify(
                "sphinx/clientDestroyed",
                ClientDestroyedNotification(id=previous_client.id),
            )
            self.server.run_task(previous_client.stop())

        resolved = config.resolve(uri, self.server.workspace, self.logger)
        if resolved is None:
            self.clients[event.scope] = None
            return

        self.clients[event.scope] = client = self.client_factory(self, resolved)
        client.add_listener("state-change", partial(self._on_state_change, event.scope))

        self.server.protocol.notify(
            "sphinx/clientCreated",
            ClientCreatedNotification(id=client.id, scope=event.scope, config=resolved),
        )
        self.logger.debug("Client created for scope %s", event.scope)

        # Start the client
        await client

    def _on_state_change(
        self,
        scope: str,
        client: SphinxClient,
        old_state: ClientState,
        new_state: ClientState,
    ):
        """React to state changes in the client."""

        if old_state == ClientState.Starting and new_state == ClientState.Running:
            if (sphinx_info := client.sphinx_info) is not None:
                self.project_manager.register_project(scope, client.db)
                self._events.trigger("app-created", client)
                self.server.protocol.notify(
                    "sphinx/appCreated",
                    AppCreatedNotification(id=client.id, application=sphinx_info),
                )

        if old_state == ClientState.Building:
            self.stop_progress(client)

        if new_state == ClientState.Building:
            self.server.run_task(self.start_progress(client))

        if new_state == ClientState.Errored:
            error = ""
            detail = ""

            if (exc := getattr(client, "exception", None)) is not None:
                error = f"{type(exc).__name__}: {exc}"
                detail = "".join(
                    traceback.format_exception(type(exc), exc, exc.__traceback__)
                )

            self.server.window_show_message(
                lsp.ShowMessageParams(message=error, type=lsp.MessageType.Error)
            )
            self.server.protocol.notify(
                "sphinx/clientErrored",
                ClientErroredNotification(id=client.id, error=error, detail=detail),
            )

    async def start_progress(self, client: SphinxClient):
        """Start reporting work done progress for the given client."""

        # Make sure any existing progress tokens are cleaned up
        if (existing := self._progress_tokens.get(client.id)) is not None:
            self.logger.warning("Overwriting existing progress token: %r!", existing)
            self.stop_progress(client)

        token = str(uuid.uuid4())
        self.logger.debug("Starting progress: '%s'", token)

        try:
            await self.server.work_done_progress.create_async(token)
        except Exception as exc:
            self.logger.debug("Unable to create progress token: %s", exc)
            return

        self._progress_tokens[client.id] = token
        self.server.work_done_progress.begin(
            token,
            lsp.WorkDoneProgressBegin(title="sphinx-build", cancellable=False),
        )

    def stop_progress(self, client: SphinxClient):
        if (token := self._progress_tokens.pop(client.id, None)) is None:
            return

        self.logger.debug("Ending progress: %r", token)
        self.server.work_done_progress.end(
            token, lsp.WorkDoneProgressEnd(message="Finished")
        )

    def report_progress(self, client: SphinxClient, progress: types.ProgressParams):
        """Report progress done for the given client."""

        if client.state not in {ClientState.Running, ClientState.Building}:
            return

        if (token := self._progress_tokens.get(client.id, None)) is None:
            return

        self.server.work_done_progress.report(
            token,
            lsp.WorkDoneProgressReport(
                message=progress.message,
                percentage=progress.percentage,
                cancellable=False,
            ),
        )


def _structure_list_or_subprocess(obj, _):
    """Structure hook that can automatically upgrade a simple python command e.g.
    ``["/bin/python"]`` to a SubProcess instance."""
    if isinstance(obj, list):
        return SubProcess(command=obj)

    return SubProcess(**obj)



# Test cases originally based on:
# https://github.com/microsoft/vscode/blob/5653420433692dc4269ad39adbc143e3438af179/src/vs/base/test/common/uri.test.ts
from __future__ import annotations

import os.path
import pathlib
from typing import Any

import pytest

from esbonio.sphinx_agent.types import IS_WIN
from esbonio.sphinx_agent.types import Uri


def test_uri_is_pathlike():
    """Ensure that a Uri implements the ``os.PathLike`` protocol and that it works as
    expected."""

    uri = Uri.for_file(__file__)
    assert os.fspath(uri) == os.fspath(__file__)
    assert os.path.normpath(os.path.join(uri, "..")) == os.path.normpath(
        os.path.join(__file__, "..")
    )
    assert pathlib.Path(uri) == pathlib.Path(__file__)


def test_uri_hashable():
    """Ensure that the Uri class is hashable, so that it can be used as dictionary
    keys etc."""

    uri = Uri.for_file("file.txt")
    assert hash(uri) != 0


@pytest.mark.parametrize(
    "a,b,expected",
    [
        (Uri.parse("file:///a.txt"), 1, False),
        (Uri.create(scheme="file"), Uri.create(scheme="file"), True),
        (Uri.create(scheme="file"), Uri.create(scheme="http"), False),
        (
            Uri.create(scheme="file", query="a"),
            Uri.create(scheme="file", query="a"),
            True,
        ),
        (
            Uri.create(scheme="file", query="a"),
            Uri.create(scheme="file", query="b"),
            False,
        ),
        (
            Uri.create(scheme="file", authority="a"),
            Uri.create(scheme="file", authority="a"),
            True,
        ),
        (
            Uri.create(scheme="file", authority="a"),
            Uri.create(scheme="file", authority="b"),
            False,
        ),
        (
            Uri.create(scheme="file", fragment="a"),
            Uri.create(scheme="file", fragment="a"),
            True,
        ),
        (
            Uri.create(scheme="file", fragment="a"),
            Uri.create(scheme="file", fragment="b"),
            False,
        ),
        (
            Uri.create(scheme="http", path="a/b"),
            Uri.create(scheme="http", path="a/b"),
            True,
        ),
        (
            Uri.create(scheme="http", path="a/b"),
            Uri.create(scheme="http", path="a/B"),
            False,
        ),
        pytest.param(
            Uri.create(scheme="file", path="a/b"),
            Uri.create(scheme="file", path="a/B"),
            False,
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            # Filepaths on Windows are case insensitive
            Uri.create(scheme="file", path="a/b"),
            Uri.create(scheme="file", path="a/B"),
            True,
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
    ],
)
def test_uri_eq(a: Uri, b: Any, expected: bool):
    """Ensure that we have implemented ``__eq__`` for Uris correctly."""
    assert (a == b) is expected


@pytest.mark.parametrize(
    "path, expected",
    [
        ("c:/win/path", "file:///c%3A/win/path"),
        ("C:/win/path", "file:///c%3A/win/path"),
        ("c:/win/path/", "file:///c%3A/win/path/"),
        ("/c:/win/path", "file:///c%3A/win/path"),
        pytest.param(
            r"c:\win\path",
            "file:///c%3A/win/path",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            r"c:\win/path",
            "file:///c%3A/win/path",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            r"c:\win\path",
            "file:///c%3A%5Cwin%5Cpath",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            r"c:\win/path",
            "file:///c%3A%5Cwin/path",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A Windows"),
        ),
    ],
)
def test_file_uri_to_string(path: str, expected: str):
    """Ensure that we can convert ``file://`` based uris to strings correctly."""
    assert str(Uri.for_file(path)) == expected


@pytest.mark.parametrize(
    "path, expected",
    [
        pytest.param(
            "c:/win/path",
            "c:/win/path",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "c:/win/path/",
            "c:/win/path/",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "C:/win/path",
            "c:/win/path",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "/c:/win/path",
            "c:/win/path",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "./c/win/path",
            "/./c/win/path",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        # -------------------
        pytest.param(
            r"c:\win\path",
            r"c:\win\path",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            r"c:\win/path",
            r"c:\win\path",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "c:/win/path",
            r"c:\win\path",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "c:/win/path/",
            "c:\\win\\path\\",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "C:/win/path",
            r"c:\win\path",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "/c:/win/path",
            r"c:\win\path",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "./c/win/path",
            r"\.\c\win\path",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
    ],
)
def test_file_uri_fs_path(path: str, expected: str):
    """Ensure that we can compute an fs path from a uri correctly."""
    assert Uri.for_file(path).fs_path == expected


@pytest.mark.parametrize(
    "src, expected",
    [
        pytest.param(
            "file:///c:/alex.txt",
            "c:/alex.txt",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "file://monacotools/folder/isi.txt",
            "//monacotools/folder/isi.txt",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "file://monacotools1/certificates/SSL/",
            "//monacotools1/certificates/SSL/",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "file:///c:/Source/Z%C3%BCrich%20or%20Zurich%20(%CB%88zj%CA%8A%C9%99r%C9%AAk,"
            "/Code/resources/app/plugins",
            "c:/Source/Zürich or Zurich (ˈzjʊərɪk,/Code/resources/app/plugins",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        # -------------------
        pytest.param(
            "file:///c:/alex.txt",
            r"c:\alex.txt",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "file://monacotools/folder/isi.txt",
            r"\\monacotools\folder\isi.txt",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "file://monacotools1/certificates/SSL/",
            "\\\\monacotools1\\certificates\\SSL\\",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "file:///c:/Source/Z%C3%BCrich%20or%20Zurich%20(%CB%88zj%CA%8A%C9%99r%C9%AAk,"
            "/Code/resources/app/plugins",
            r"c:\Source\Zürich or Zurich (ˈzjʊərɪk,\Code\resources\app\plugins",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
    ],
)
def test_uri_from_fs_path(src: str, expected: str):
    """Ensure that we can compute an ``fs_path`` correctly, create a uri from it and
    have it be equivalent to the original uri."""

    uri = Uri.parse(src)
    assert uri.fs_path == expected

    uri2 = Uri.for_file(uri.fs_path)
    assert uri2.fs_path == expected
    assert str(uri) == str(uri2)


def test_fs_path_with_no_path():
    """Ensure that the ``fs_path`` property behaves as expected when the uri contains
    no path."""
    uri = Uri.parse("file://%2Fhome%2Fticino%2Fdesktop%2Fcpluscplus%2Ftest.cpp")
    assert uri.authority == "/home/ticino/desktop/cpluscplus/test.cpp"
    assert uri.path == "/"

    expected = "\\" if IS_WIN else "/"
    assert uri.fs_path == expected


@pytest.mark.parametrize(
    "uri, expected",
    [
        (
            Uri.create(scheme="http", authority="www.example.com", path="/my/path"),
            "http://www.example.com/my/path",
        ),
        (
            Uri.create(scheme="http", authority="www.EXAMPLE.com", path="/my/path"),
            "http://www.example.com/my/path",
        ),
        (Uri.create(scheme="http", path="my/path"), "http:/my/path"),
        (Uri.create(scheme="http", path="/my/path"), "http:/my/path"),
        (
            Uri.create(scheme="http", authority="example.com", query="test=true"),
            "http://example.com/?test%3Dtrue",
        ),
        (
            Uri.create(scheme="http", authority="example.com", fragment="test=true"),
            "http://example.com/#test%3Dtrue",
        ),
        (Uri.parse("stuff:?qüery"), "stuff:?q%C3%BCery"),
        (Uri.parse("file://sh%c3%a4res/path"), "file://sh%C3%A4res/path"),
        (
            Uri.parse("untitled:c:/Users/jreiken/Code/abc.txt"),
            "untitled:c%3A/Users/jreiken/Code/abc.txt",
        ),
        (
            Uri.parse("untitled:C:/Users/jreiken/Code/abc.txt"),
            "untitled:c%3A/Users/jreiken/Code/abc.txt",
        ),
        (
            Uri.for_file(
                "/Users/jrieken/Code/_samples/18500/Mödel + Other Thîngß/model.js"
            ),
            "file:///Users/jrieken/Code/_samples/18500/M%C3%B6del%20%2B%20Other%20Th%C3%AEng%C3%9F/model.js",
        ),
        (Uri.parse("http://localhost:8080/far"), "http://localhost:8080/far"),
        (Uri.parse("http://löcalhost:8080/far"), "http://l%C3%B6calhost:8080/far"),
        (Uri.parse("http://foo:bar@localhost/far"), "http://foo:bar@localhost/far"),
        (Uri.parse("http://foo@localhost/far"), "http://foo@localhost/far"),
        (
            Uri.parse("http://foo:bAr@localhost:8080/far"),
            "http://foo:bAr@localhost:8080/far",
        ),
        (Uri.parse("http://foo@localhost:8080/far"), "http://foo@localhost:8080/far"),
        (
            Uri.parse("http://föö:bör@löcalhost:8080/far"),
            "http://f%C3%B6%C3%B6:b%C3%B6r@l%C3%B6calhost:8080/far",
        ),
        (
            Uri.parse("https://go.microsoft.com/fwlink/?LinkId=518008"),
            "https://go.microsoft.com/fwlink/?LinkId%3D518008",
        ),
        (
            Uri.parse("https://go.microsoft.com/fwlink/?LinkId=518008&foö&ké¥=üü"),
            "https://go.microsoft.com/fwlink/?LinkId%3D518008%26fo%C3%B6%26k%C3%A9%C2%A5%3D%C3%BC%C3%BC",
        ),
    ],
)
def test_uri_to_string_with_encoding(uri: Uri, expected: str):
    """Ensure that we can convert a uri to a string correctly with encoding"""
    assert str(uri) == expected


@pytest.mark.parametrize(
    "uri, expected",
    [
        (
            Uri.create(scheme="http", authority="example.com", query="test=true"),
            "http://example.com/?test=true",
        ),
        (
            Uri.create(scheme="http", authority="example.com", fragment="test=true"),
            "http://example.com/#test=true",
        ),
        (
            Uri.create(scheme="http", path="/api/files/test.me", query="t=1234"),
            "http:/api/files/test.me?t=1234",
        ),
        (
            Uri.parse("https://go.microsoft.com/fwlink/?LinkId=518008"),
            "https://go.microsoft.com/fwlink/?LinkId=518008",
        ),
        (
            Uri.parse("https://go.microsoft.com/fwlink/?LinkId=518008&foö&ké¥=üü"),
            "https://go.microsoft.com/fwlink/?LinkId=518008&foö&ké¥=üü",
        ),
        (
            Uri.parse("https://twitter.com/search?src=typd&q=%23tag"),
            "https://twitter.com/search?src=typd&q=%23tag",
        ),
    ],
)
def test_uri_to_string_without_encoding(uri: Uri, expected: str):
    """Ensure that we can convert a uri to a string correctly with encoding"""
    assert uri.as_string(encode=False) == expected


def test_uri_parse_encode():
    uri = Uri.parse("file://shares/pröjects/c%23/#l12")
    assert uri.authority == "shares"
    assert uri.path == "/pröjects/c#/"
    assert uri.fragment == "l12"
    assert str(uri) == "file://shares/pr%C3%B6jects/c%23/#l12"
    assert uri.as_string(encode=False) == "file://shares/pröjects/c%23/#l12"

    uri2 = Uri.parse(str(uri))
    uri3 = Uri.parse(uri.as_string(encode=False))
    assert uri2.authority == uri3.authority
    assert uri2.path == uri3.path
    assert uri2.query == uri3.query
    assert uri2.fragment == uri3.fragment


@pytest.mark.parametrize(
    "before, args, expected",
    [
        (
            Uri.parse("before:some/file/path"),
            dict(scheme="after"),
            "after:some/file/path",
        ),
        (
            Uri.create(scheme="s"),
            dict(scheme="http", path="/api/files/test.me", query="t=1234"),
            "http:/api/files/test.me?t%3D1234",
        ),
        (
            Uri.create(scheme="s"),
            dict(
                scheme="http", authority="", path="/api/files/test.me", query="t=1234"
            ),
            "http:/api/files/test.me?t%3D1234",
        ),
        (
            Uri.create(scheme="s"),
            dict(
                scheme="https", authority="", path="/api/files/test.me", query="t=1234"
            ),
            "https:/api/files/test.me?t%3D1234",
        ),
        (
            Uri.create(scheme="s"),
            dict(
                scheme="HTTP", authority="", path="/api/files/test.me", query="t=1234"
            ),
            "HTTP:/api/files/test.me?t%3D1234",
        ),
        (
            Uri.create(scheme="s"),
            dict(
                scheme="HTTPS", authority="", path="/api/files/test.me", query="t=1234"
            ),
            "HTTPS:/api/files/test.me?t%3D1234",
        ),
        (
            Uri.create(scheme="s"),
            dict(scheme="boo", authority="", path="/api/files/test.me", query="t=1234"),
            "boo:/api/files/test.me?t%3D1234",
        ),
        (Uri.parse("scheme://authority/path"), dict(authority=""), "scheme:/path"),
        (Uri.parse("scheme://authority/path"), dict(authority=None), "scheme:/path"),
        (Uri.parse("scheme://authority/path"), dict(path=""), "scheme://authority"),
        (Uri.parse("scheme://authority/path"), dict(path=None), "scheme://authority"),
    ],
)
def test_uri_where(before: Uri, args: dict[str, str], expected: str):
    """Ensure that we can transform uris correctly."""
    assert str(before.where(**args)) == expected


@pytest.mark.parametrize(
    "before, args, message",
    [
        (Uri.parse("foo:bar/path"), dict(scheme=None), "URIs must have a scheme"),
        (Uri.parse("foo:bar/path"), dict(scheme="fai:l"), "Invalid scheme"),
        (Uri.parse("foo:bar/path"), dict(scheme="fäil"), "Invalid scheme"),
        (
            Uri.parse("foo:bar/path"),
            dict(authority="fail"),
            "Paths with an authority must start with a slash '/'",
        ),
        (
            Uri.parse("foo:bar/path"),
            dict(path="//fail"),
            "Paths without an authority cannot start with two slashes '//'",
        ),
    ],
)
def test_uri_where_validation(before: Uri, args: dict[str, str], message: str):
    """Ensure that transformed uris are validated"""

    with pytest.raises(ValueError) as e:
        before.where(**args)

    assert message in str(e)


@pytest.mark.parametrize(
    "uri, expected",
    [
        (
            "http:/api/files/test.me?t=1234",
            Uri.create(scheme="http", path="/api/files/test.me", query="t=1234"),
        ),
        (
            "http://api/files/test.me?t=1234",
            Uri.create(
                scheme="http", authority="api", path="/files/test.me", query="t=1234"
            ),
        ),
        ("file:///c:/test/me", Uri.create(scheme="file", path="/c:/test/me")),
        (
            "file://shares/files/c%23/p.cs",
            Uri.create(scheme="file", authority="shares", path="/files/c#/p.cs"),
        ),
        (
            "file:///c:/Source/Z%C3%BCrich%20or%20Zurich%20(%CB%88zj%CA%8A%C9%99r%C9%AAk,"
            "/Code/resources/app/plugins/c%23/plugin.json",
            Uri.create(
                scheme="file",
                path="/c:/Source/Zürich or Zurich (ˈzjʊərɪk,/Code/resources/app/plugins/c#/plugin.json",
            ),
        ),
        ("file:///c:/test %25/path", Uri.create(scheme="file", path="/c:/test %/path")),
        ("inmemory:", Uri.create(scheme="inmemory")),
        ("foo:api/files/test", Uri.create(scheme="foo", path="api/files/test")),
        ("file:?q", Uri.create(scheme="file", query="q")),
        ("file:#d", Uri.create(scheme="file", fragment="d")),
        ("f3ile:#d", Uri.create(scheme="f3ile", fragment="d")),
        ("foo+bar:path", Uri.create(scheme="foo+bar", path="path")),
        ("foo-bar:path", Uri.create(scheme="foo-bar", path="path")),
        ("foo.bar:path", Uri.create(scheme="foo.bar", path="path")),
    ],
)
def test_uri_parse(uri: str, expected: Uri):
    """Ensure that we can parse uris correctly."""
    assert Uri.parse(uri) == expected


def test_uri_parse_drive_letter():
    src = "file:///_:/path"
    uri = Uri.parse(src)

    assert uri == Uri.create(scheme="file", path="/_:/path")
    assert uri.fs_path == r"\_:\path" if IS_WIN else "/_:/path"


def test_uri_parse_validation():
    """Ensure that we validate uris when parsing"""

    with pytest.raises(ValueError):
        Uri.parse("file:////shares/files/p.cs")


@pytest.mark.skipif(not IS_WIN, reason="Windows only")
@pytest.mark.parametrize(
    "path, expected_uri, expected_str",
    [
        (
            r"c:\test\drive",
            Uri.create(scheme="file", path="/c:/test/drive"),
            "file:///c%3A/test/drive",
        ),
        (
            r"\\shäres\path\c#\plugin.json",
            Uri.create(scheme="file", authority="shäres", path="/path/c#/plugin.json"),
            "file://sh%C3%A4res/path/c%23/plugin.json",
        ),
        (
            r"\\localhost\c$\GitDevelopment\express",
            Uri.create(
                scheme="file", authority="localhost", path="/c$/GitDevelopment/express"
            ),
            "file://localhost/c%24/GitDevelopment/express",
        ),
        (
            r"c:\test with %\path",
            Uri.create(scheme="file", path="/c:/test with %/path"),
            "file:///c%3A/test%20with%20%25/path",
        ),
        (
            r"c:\test with %25\path",
            Uri.create(scheme="file", path="/c:/test with %25/path"),
            "file:///c%3A/test%20with%20%2525/path",
        ),
        (
            r"c:\test with %25\c#code",
            Uri.create(scheme="file", path="/c:/test with %25/c#code"),
            "file:///c%3A/test%20with%20%2525/c%23code",
        ),
        (
            r"\\shares",
            Uri.create(scheme="file", authority="shares", path="/"),
            "file://shares/",
        ),
        (
            "\\\\shares\\",
            Uri.create(scheme="file", authority="shares", path="/"),
            "file://shares/",
        ),
    ],
)
def test_uri_for_file(path: str, expected_uri: Uri, expected_str: str):
    """Ensure that we can construct a uri for a filepath correctly."""
    uri = Uri.for_file(path)

    assert uri == expected_uri
    assert str(uri) == expected_str


@pytest.mark.parametrize(
    "base, extra, expected",
    [
        ("file:///foo/", "../../bazz", "file:///bazz"),
        ("file:///foo", "../../bazz", "file:///bazz"),
        ("file:///foo/bar", "./bazz", "file:///foo/bar/bazz"),
        ("file:///foo/bar", "bazz", "file:///foo/bar/bazz"),
        ("file:", "bazz", "file:///bazz"),
        ("http://domain", "bazz", "http://domain/bazz"),
        ("https://domain", "bazz", "https://domain/bazz"),
        ("http:", "bazz", "http:/bazz"),
        ("https:", "bazz", "https:/bazz"),
        ("foo:/", "bazz", "foo:/bazz"),
        ("foo://bar/", "bazz", "foo://bar/bazz"),
        pytest.param(
            "file:///c:/foo/",
            "../../bazz",
            "file:///bazz",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "file://server/share/c:/",
            "../../bazz",
            "file://server/bazz",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "file://server/share/c:",
            "../../bazz",
            "file://server/bazz",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "file://ser/foo/",
            "../../bazz",
            "file://ser/bazz",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "file://ser/foo",
            "../../bazz",
            "file://ser/bazz",
            marks=pytest.mark.skipif(IS_WIN, reason="N/A for Windows"),
        ),
        pytest.param(
            "file:///c:/foo/",
            "../../bazz",
            "file:///c:/bazz",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "file://server/share/c:/",
            "../../bazz",
            "file://server/share/bazz",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "file://server/share/c:",
            "../../bazz",
            "file://server/share/bazz",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "file://ser/foo/",
            "../../bazz",
            "file://ser/foo/bazz",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "file://ser/foo",
            "../../bazz",
            "file://ser/foo/bazz",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
        pytest.param(
            "file:///c:/foo/bar",
            "./other/foo.img",
            "file:///c:/foo/bar/other/foo.img",
            marks=pytest.mark.skipif(not IS_WIN, reason="Windows only"),
        ),
    ],
)
def test_uri_join_path(base: str, extra: str, expected):
    """Ensure that we can join uri paths correctly."""

    uri = Uri.parse(base).join(extra)
    assert uri.as_string(encode=False) == expected

    uri2 = Uri.parse(base) / extra
    assert uri2.as_string(encode=False) == expected


@pytest.mark.parametrize(
    "base, extra",
    [("foo:", "bazz"), ("foo://bar", "bazz")],
)
def test_uri_join_path_error(base: str, extra: str):
    """Ensure that we can join uri paths correctly."""

    with pytest.raises(ValueError) as e:
        Uri.parse(base).join(extra)

    assert "has no path" in str(e)

    with pytest.raises(ValueError) as e:
        _ = Uri.parse(base) / extra

    assert "has no path" in str(e)



"""Technically, not a true end-to-end test as we're inspecting the internals of the server.
But since it depends on the sphinx agent, we may as well put it here."""

from __future__ import annotations

import asyncio
import io
import pathlib
import sys
import typing

import pytest
import pytest_asyncio
from lsprotocol import types as lsp
from pygls import IS_WIN

from esbonio.server import EsbonioLanguageServer
from esbonio.server import Uri
from esbonio.server import create_language_server
from esbonio.server.features.project_manager import ProjectManager
from esbonio.server.features.sphinx_manager import ClientState
from esbonio.server.features.sphinx_manager import SphinxManager
from esbonio.server.features.sphinx_manager import make_subprocess_sphinx_client

if typing.TYPE_CHECKING:
    from collections.abc import Coroutine
    from typing import Any
    from typing import Protocol

    class ServerManagerFactory(Protocol):
        def __call__(
            self,
            *,
            init_options: dict[str, Any] | None = None,
            workspace_config: dict[str, Any] | None = None,
        ) -> Coroutine[None, None, tuple[EsbonioLanguageServer, SphinxManager]]: ...


@pytest.fixture
def demo_workspace(uri_for):
    return uri_for("workspaces", "demo")


@pytest.fixture
def docs_workspace(uri_for):
    return uri_for("..", "..", "..", "docs")


@pytest_asyncio.fixture()
async def server_manager(demo_workspace: Uri, docs_workspace):
    """An instance of the language server and sphinx manager to use for the test."""

    esbonio = create_language_server(EsbonioLanguageServer, [])
    esbonio.protocol.set_writer(io.BytesIO())

    project_manager = ProjectManager(esbonio)
    esbonio.add_feature(project_manager)

    sphinx_manager = SphinxManager(
        make_subprocess_sphinx_client, project_manager, esbonio
    )
    esbonio.add_feature(sphinx_manager)

    async def initialize(
        *,
        init_options: dict[str, Any] | None = None,
        workspace_config: dict[str, Any] | None = None,
    ):
        # Initialize the server.
        esbonio.protocol.handle_message(
            lsp.InitializeRequest(
                id=1,
                params=lsp.InitializeParams(
                    capabilities=lsp.ClientCapabilities(),
                    initialization_options=init_options,
                    workspace_folders=[
                        lsp.WorkspaceFolder(uri=str(demo_workspace), name="demo"),
                        lsp.WorkspaceFolder(uri=str(docs_workspace), name="docs"),
                    ],
                ),
            )
        )

        esbonio.protocol.handle_message(
            lsp.InitializedNotification(params=lsp.InitializedParams())
        )

        # Ensure that the server is ready
        retry = 10
        while (not esbonio.ready.done()) and retry > 0:
            await asyncio.sleep(0.5)
            retry -= 1

        if not esbonio.ready.done():
            raise RuntimeError("Server did not initialize")

        # Hack to set the workspace configuration we require.
        esbonio.configuration._workspace_config = workspace_config or {}

        return esbonio, sphinx_manager

    yield initialize

    await sphinx_manager.shutdown(None)


@pytest.mark.asyncio
async def test_get_client(
    server_manager: ServerManagerFactory, demo_workspace: Uri, tmp_path: pathlib.Path
):
    """Ensure that we can create a SphinxClient correctly."""

    _, manager = await server_manager(
        workspace_config={
            str(demo_workspace): dict(
                esbonio=dict(
                    sphinx=dict(
                        pythonCommand=[sys.executable],
                        buildCommand=[
                            "sphinx-build",
                            "-M",
                            "dirhtml",
                            ".",
                            str(tmp_path),
                        ],
                        configOverrides={
                            "html_theme": "alabaster",
                            "html_theme_options": {},
                        },
                    ),
                ),
            ),
        },
    )

    result = await manager.get_client(demo_workspace / "index.rst")
    # At least for now, the first call to get_client will not return a client
    # but will instead "prime the system" ready to return a client the next
    # time it is called.
    assert result is None

    # Give the async tasks chance to complete.
    await asyncio.sleep(0.5)

    # A client should have been created and started
    client = manager.clients[str(demo_workspace)]

    assert client is not None
    assert client.state in {ClientState.Starting, ClientState.Running}

    # Now when we ask for the client, the client should be started and we should
    # get back the same instance
    result = await manager.get_client(demo_workspace / "index.rst")

    assert result is client
    assert client.state == ClientState.Running

    # And that the client initialized correctly
    assert client.builder == "dirhtml"
    assert client.src_uri == demo_workspace
    assert client.conf_uri == demo_workspace
    assert client.build_uri == Uri.for_file(tmp_path / "dirhtml")


@pytest.mark.asyncio
async def test_get_client_with_error(
    server_manager: ServerManagerFactory, demo_workspace: Uri
):
    """Ensure that we correctly handle the case where there is an error with the client."""

    _, manager = await server_manager(
        workspace_config={
            str(demo_workspace): dict(
                esbonio=dict(
                    sphinx=dict(
                        pythonCommand=["/not/a/real/env/python"],
                    ),
                ),
            ),
        },
    )

    result = await manager.get_client(demo_workspace / "index.rst")
    # At least for now, the first call to get_client will not return a client
    # but will instead "prime the system" ready to return a client the next
    # time it is called.
    assert result is None

    # Give the async tasks chance to complete.
    await asyncio.sleep(0.5)

    # A client should have been created and started
    client = manager.clients[str(demo_workspace)]

    assert client is not None
    assert client.state in {ClientState.Starting, ClientState.Errored}

    # Now when we ask for the client, the client should be started and we should
    # get back the same instance
    result = await manager.get_client(demo_workspace / "index.rst")

    assert result is client
    assert client.state == ClientState.Errored

    if IS_WIN:
        message = "The system cannot find the file specified"
    else:
        message = "No such file or directory"

    assert message in str(client.exception)

    # Finally, if we request another uri from the same project we should get back
    # the same client instance - even though it failed to start.
    result = await manager.get_client(demo_workspace / "demo_myst.md")
    assert result is client


@pytest.mark.asyncio
async def test_get_client_with_many_uris(
    server_manager: ServerManagerFactory, demo_workspace: Uri, tmp_path: pathlib.Path
):
    """Ensure that when called in rapid succession, with many uris we only create a
    single client instance."""

    _, manager = await server_manager(
        workspace_config={
            str(demo_workspace): dict(
                esbonio=dict(
                    sphinx=dict(
                        pythonCommand=[sys.executable],
                        buildCommand=[
                            "sphinx-build",
                            "-M",
                            "dirhtml",
                            ".",
                            str(tmp_path),
                        ],
                        configOverrides={
                            "html_theme": "alabaster",
                            "html_theme_options": {},
                        },
                    ),
                ),
            ),
        },
    )

    src_uris = [Uri.for_file(f) for f in pathlib.Path(demo_workspace).glob("**/*.rst")]
    coros = [manager.get_client(s) for s in src_uris]

    # As with the other cases, this should only "prime" the system
    result = await asyncio.gather(*coros)
    assert all([r is None for r in result])

    # There should only have been one client created (in addition to the 'dummy' global
    # scoped client)
    assert len(manager.clients) == 2

    client = manager.clients[str(demo_workspace)]
    assert client is not None
    assert client.state == ClientState.Starting

    # Now if we do the same again we should get the same client instance for each case.
    coros = [manager.get_client(s) for s in src_uris]
    result = await asyncio.gather(*coros)

    assert all([r is client for r in result])

    client = result[0]
    assert client.state == ClientState.Running

    # And that the client initialized correctly
    assert client.builder == "dirhtml"
    assert client.src_uri == demo_workspace
    assert client.conf_uri == demo_workspace
    assert client.build_uri == Uri.for_file(tmp_path / "dirhtml")


@pytest.mark.asyncio
async def test_get_client_with_many_uris_in_many_projects(
    server_manager: ServerManagerFactory,
    demo_workspace: Uri,
    docs_workspace: Uri,
    tmp_path: pathlib.Path,
):
    """Ensure that when called in rapid succession, with many uris we only create a
    single client instance for each project."""

    _, manager = await server_manager(
        workspace_config={
            str(demo_workspace): dict(
                esbonio=dict(
                    sphinx=dict(
                        pythonCommand=[sys.executable],
                        buildCommand=[
                            "sphinx-build",
                            "-M",
                            "dirhtml",
                            ".",
                            str(tmp_path / "demo"),
                        ],
                        configOverrides={
                            "html_theme": "alabaster",
                            "html_theme_options": {},
                        },
                    ),
                ),
            ),
            str(docs_workspace): dict(
                esbonio=dict(
                    sphinx=dict(
                        pythonCommand=[sys.executable],
                        buildCommand=[
                            "sphinx-build",
                            "-M",
                            "dirhtml",
                            ".",
                            str(tmp_path / "docs"),
                        ],
                        configOverrides={
                            "html_theme": "alabaster",
                            "html_theme_options": {},
                        },
                    ),
                ),
            ),
        },
    )

    src_uris = [Uri.for_file(f) for f in pathlib.Path(demo_workspace).glob("**/*.rst")]
    src_uris += [Uri.for_file(f) for f in pathlib.Path(docs_workspace).glob("**/*.rst")]
    coros = [manager.get_client(s) for s in src_uris]

    # As with the other cases, this should only "prime" the system
    result = await asyncio.gather(*coros)
    assert all([r is None for r in result])

    # There should only have been one client created for each project (in addition to
    # the 'dummy' global scoped client)
    assert len(manager.clients) == 3

    demo_client = manager.clients[str(demo_workspace)]
    assert demo_client is not None
    assert demo_client.state == ClientState.Starting

    docs_client = manager.clients[str(docs_workspace)]
    assert docs_client is not None
    assert docs_client.state == ClientState.Starting

    # Now if we do the same again we should get the same client instance for each case.
    coros = [manager.get_client(s) for s in src_uris]
    result = await asyncio.gather(*coros)

    assert all([(r is demo_client) or (r is docs_client) for r in result])

    assert demo_client.state == ClientState.Running

    # When run in CI, the docs won't have all the required dependencies available.
    assert docs_client.state in {ClientState.Running, ClientState.Errored}


@pytest.mark.asyncio
async def test_updated_config(
    server_manager: ServerManagerFactory, demo_workspace: Uri, tmp_path: pathlib.Path
):
    """Ensure that when the configuration affecting a Sphinx configuration is changed,
    the SphinxClient is recreated."""

    server, manager = await server_manager(
        workspace_config={
            str(demo_workspace): dict(
                esbonio=dict(
                    sphinx=dict(
                        pythonCommand=[sys.executable],
                        buildCommand=[
                            "sphinx-build",
                            "-M",
                            "dirhtml",
                            ".",
                            str(tmp_path),
                        ],
                        configOverrides={
                            "html_theme": "alabaster",
                            "html_theme_options": {},
                        },
                    ),
                ),
            ),
        }
    )

    result = await manager.get_client(demo_workspace / "index.rst")
    # At least for now, the first call to get_client will not return a client
    # but will instead "prime the system" ready to return a client the next
    # time it is called.
    assert result is None

    # Give the async tasks chance to complete.
    await asyncio.sleep(0.5)

    # A client should have been created and started
    client = manager.clients[str(demo_workspace)]

    assert client is not None
    assert client.state in {ClientState.Starting, ClientState.Running}

    # Now when we ask for the client, the client should be started and we should
    # get back the same instance
    result = await manager.get_client(demo_workspace / "index.rst")

    assert result is client
    assert client.state == ClientState.Running

    # And that the client initialized correctly
    assert client.builder == "dirhtml"

    # Now update the configuration
    wkspace_config = server.configuration._workspace_config[str(demo_workspace)]
    wkspace_config["esbonio"]["sphinx"]["buildCommand"] = [
        "sphinx-build",
        "-M",
        "html",
        ".",
        str(tmp_path),
    ]
    server.configuration._notify_subscriptions()

    # Give the async tasks chance to complete.
    await asyncio.sleep(0.5)

    # A new client should have been created, started and be using the new config
    new_client = manager.clients[str(demo_workspace)]

    assert new_client is not client
    assert new_client.state in {ClientState.Starting, ClientState.Running}

    # Ensure that the client has finished starting
    await new_client
    assert new_client.builder == "html"

    # The old client should have been stopped
    assert client.state == ClientState.Exited



from __future__ import annotations

import asyncio
import inspect
import logging
import traceback
import typing
from functools import partial

if typing.TYPE_CHECKING:
    from typing import Any


class EventSource:
    """Simple component for emitting events."""

    # TODO: It might be nice to do some fancy typing here so that type checkers
    # etc know which events are possible etc.

    def __init__(self, logger: logging.Logger | None = None):
        self.logger = logger or logging.getLogger(__name__)
        """The logging instance to use."""

        self.handlers: dict[str, set] = {}
        """Collection of handlers for various events."""

        self._tasks: set[asyncio.Task] = set()
        """Holds tasks that are currently executing an async event handler."""

    def add_listener(self, event: str, handler):
        """Add a listener for the given event name."""
        self.handlers.setdefault(event, set()).add(handler)

    def _finish_task(self, event: str, listener_name: str, task: asyncio.Task[Any]):
        """Cleanup a finished task."""
        self._tasks.discard(task)

        if (exc := task.exception()) is not None:
            self.logger.error(
                "Error in '%s' async handler '%s'\n%s",
                event,
                listener_name,
                "".join(traceback.format_exception(type(exc), exc, exc.__traceback__)),
            )

    def trigger(self, event: str, *args, **kwargs):
        """Trigger the event with the given name."""

        for listener in self.handlers.get(event, set()):
            listener_name = f"{listener}"

            try:
                res = listener(*args, **kwargs)

                # Event listeners may be async
                if inspect.iscoroutine(res):
                    task = asyncio.create_task(res)
                    task.add_done_callback(
                        partial(self._finish_task, event, listener_name)
                    )

                    self._tasks.add(task)

            except Exception:
                self.logger.error(
                    "Error in '%s' handler '%s'", event, listener_name, exc_info=True
                )



from __future__ import annotations

import pathlib
import sqlite3
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Literal


class Database:
    @dataclass
    class Column:
        name: str
        dtype: str
        notnull: bool = field(default=False)
        default: Any | None = field(default=None)
        pk: int = field(default=0)

        @property
        def definition(self):
            # TODO: Is there a way to do this via a prepared statement?
            return f"{self.name} {self.dtype}"

    @dataclass
    class Table:
        name: str
        columns: list[Database.Column]

        @property
        def create_statement(self):
            """Return the SQL statement required to create this table."""
            # TODO: Is there a way to do this via a prepared statement?
            columns = ",".join([c.definition for c in self.columns])
            return "".join([f"CREATE TABLE {self.name} (", columns, ");"])

    def __init__(self, dbpath: pathlib.Path | Literal[":memory:"]):
        self.path = dbpath

        if isinstance(self.path, pathlib.Path) and not self.path.parent.exists():
            self.path.parent.mkdir(parents=True)

        self.db = sqlite3.connect(self.path)

        # Ensure that Write Ahead Logging is enabled.
        self.db.execute("PRAGMA journal_mode(WAL)")

        self._checked_tables: set[str] = set()

    def _get_table(self, name: str) -> Table | None:
        """Get the table with the given name, if it exists."""
        # TODO: SQLite does not seem to like '?' syntax in this statement...
        cursor = self.db.execute(f"PRAGMA table_info({name});")
        rows = cursor.fetchall()

        if len(rows) == 0:
            # Table does not exist
            return None

        columns = [
            self.Column(name=name, dtype=type_, notnull=notnull, default=default, pk=pk)
            for (_, name, type_, notnull, default, pk) in rows
        ]

        return self.Table(name=name, columns=columns)

    def _create_table(self, table: Table):
        """Create the given table."""
        cursor = self.db.cursor()

        # TODO: Is there a way to do this via a prepared statement?
        cursor.execute(f"DROP TABLE IF EXISTS {table.name}")
        cursor.execute(table.create_statement)
        self.db.commit()

    def clear_table(self, table: Table, **kwargs):
        """Clear the given table

        Parameters
        ----------
        kwargs
           Constraints to limit the rows that get cleared
        """

        # TODO: Is there a way to pass the table name as a '?' parameter?
        base_query = f"DELETE FROM {table.name}"  # noqa: S608
        where: list[str] = []
        parameters: list[Any] = []

        for param, value in kwargs.items():
            if value is None:
                where.append(f"{param} is null")
            else:
                where.append(f"{param} = ?")
                parameters.append(value)

        if where:
            conditions = " AND ".join(where)
            query = " ".join([base_query, "WHERE", conditions])
        else:
            query = base_query

        cursor = self.db.cursor()
        cursor.execute(query, tuple(parameters))
        self.db.commit()

    def ensure_table(self, table: Table):
        """Ensure that the given table exists in the database.

        If the table *does* exist, but has the wrong shape, it will be dropped and
        recreated.
        """
        # If we've already checked the table, then there's nothing to do
        if table.name in self._checked_tables:
            return

        if (existing := self._get_table(table.name)) is None:
            self._create_table(table)
            return

        # Are the tables compatible?
        if len(existing.columns) != len(table.columns):
            self._create_table(table)
        else:
            for existing_col, col in zip(existing.columns, table.columns, strict=False):
                if existing_col.name != col.name or existing_col.dtype != col.dtype:
                    self._create_table(table)
                    break

        self._checked_tables.add(table.name)

    def insert_values(self, table: Table, values: list[tuple]):
        """Insert the given values into the given table."""

        if len(values) == 0:
            return

        cursor = self.db.cursor()

        placeholder = "(" + ",".join(["?" for _ in range(len(values[0]))]) + ")"
        cursor.executemany(f"INSERT INTO {table.name} VALUES {placeholder}", values)  # noqa: S608
        self.db.commit()



from __future__ import annotations

import enum
import typing
from typing import Protocol

if typing.TYPE_CHECKING:
    import pathlib
    from collections.abc import Generator
    from typing import Any

    from esbonio.server import Uri
    from esbonio.sphinx_agent import types


class ClientState(enum.Enum):
    """The set of possible states the client may be in."""

    Starting = enum.auto()
    """The client is starting."""

    Restarting = enum.auto()
    """The client is restarting."""

    Running = enum.auto()
    """The client is running normally."""

    Building = enum.auto()
    """The client is currently building."""

    Errored = enum.auto()
    """The client has enountered some unrecoverable error and should not be used."""

    Exited = enum.auto()
    """The client is no longer running."""


class SphinxClient(Protocol):
    """Describes the API language features can use to inspect/manipulate a Sphinx
    application instance."""

    state: ClientState | None
    sphinx_info: types.SphinxInfo | None

    @property
    def id(self) -> str:
        """The id of the Sphinx instance."""
        ...

    @property
    def db(self) -> pathlib.Path:
        """Connection to the associated database."""
        ...

    @property
    def builder(self) -> str:
        """The name of the Sphinx builder."""
        ...

    @property
    def build_uri(self) -> Uri:
        """The URI to the Sphinx application's build dir."""
        ...

    @property
    def conf_uri(self) -> Uri:
        """The URI to the Sphinx application's conf dir."""
        ...

    @property
    def src_uri(self) -> Uri:
        """The URI to the Sphinx application's src dir."""
        ...

    def __await__(self) -> Generator[Any, None, SphinxClient]:
        """A SphinxClient should be awaitable"""
        ...

    def add_listener(self, event: str, handler): ...

    async def start(self) -> SphinxClient:
        """Start the client."""
        ...

    async def restart(self) -> SphinxClient:
        """Restart the client."""
        ...

    async def build(
        self,
        *,
        filenames: list[str] | None = None,
        force_all: bool = False,
        content_overrides: dict[str, str] | None = None,
    ) -> types.BuildResult:
        """Trigger a Sphinx build."""
        ...

    async def stop(self):
        """Stop the client."""



from __future__ import annotations

import re
import typing
from dataclasses import dataclass
from dataclasses import field
from typing import Any

from .lsp import Location

if typing.TYPE_CHECKING:
    from collections.abc import Callable
    from typing import TypeVar

    T = TypeVar("T")
    JsonLoader = Callable[[str, type[T]], T]


MYST_ROLE: re.Pattern = re.compile(
    r"""
    ([^\w`]|^\s*)                     # roles cannot be preceeded by letter chars
    (?P<role>
      {                               # roles start with a '{'
      (?P<name>[:\w+-]+)?             # roles have a name
      }?                              # roles end with a '}'
    )
    (?P<target>
      `                               # targets begin with a '`' character
      ((?P<alias>[^<`>]*?)<)?         # targets may specify an alias
      (?P<modifier>[!~])?             # targets may have a modifier
      (?P<label>[^<`>]*)?             # targets contain a label
      >?                              # labels end with a '>' when there's an alias
      `?                              # targets end with a '`' character
    )?
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse parial and complete roles.

I'm not sure if there are offical names for the components of a role, but the
language server breaks a role down into a number of parts::

                 vvvvvv label
                v modifier(optional)
               vvvvvvvv target
   {c:function}`!malloc`
   ^^^^^^^^^^^^ role
    ^^^^^^^^^^ name

The language server sometimes refers to the above as a "plain" role, in that the
role's target contains just the label of the object it is linking to. However it's
also possible to define "aliased" roles, where the link text in the final document
is overriden, for example::

                vvvvvvvvvvvvvvvvvvvvvvvv alias
                                          vvvvvv label
                                         v modifier (optional)
               vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv target
   {c:function}`used to allocate memory <~malloc>`
   ^^^^^^^^^^^^ role
    ^^^^^^^^^^ name

"""


RST_ROLE = re.compile(
    r"""
    ([^\w:]|^\s*)                     # roles cannot be preceeded by letter chars
    (?P<role>
      :                               # roles begin with a ':' character
      (?!:)                           # the next character cannot be a ':'
      ((?P<name>\w([:\w+-]*\w)?):?)?  # roles have a name
    )
    (?P<target>
      `                               # targets begin with a '`' character
      ((?P<alias>[^<`>]*?)<)?         # targets may specify an alias
      (?P<modifier>[!~])?             # targets may have a modifier
      (?P<label>[^<`>]*)?             # targets contain a label
      >?                              # labels end with a '>' when there's an alias
      `?                              # targets end with a '`' character
    )?
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse parial and complete roles.

I'm not sure if there are offical names for the components of a role, but the
language server breaks a role down into a number of parts::

                 vvvvvv label
                v modifier(optional)
               vvvvvvvv target
   :c:function:`!malloc`
   ^^^^^^^^^^^^ role
    ^^^^^^^^^^ name

The language server sometimes refers to the above as a "plain" role, in that the
role's target contains just the label of the object it is linking to. However it's
also possible to define "aliased" roles, where the link text in the final document
is overriden, for example::

                vvvvvvvvvvvvvvvvvvvvvvvv alias
                                          vvvvvv label
                                         v modifier (optional)
               vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv target
   :c:function:`used to allocate memory <~malloc>`
   ^^^^^^^^^^^^ role
    ^^^^^^^^^^ name

"""


RST_DEFAULT_ROLE = re.compile(
    r"""
    (?<![:`])
    (?P<target>
      `                               # targets begin with a '`' character
      ((?P<alias>[^<`>]*?)<)?         # targets may specify an alias
      (?P<modifier>[!~])?             # targets may have a modifier
      (?P<label>[^<`>]*)?             # targets contain a label
      >?                              # labels end with a '>' when there's an alias
      `?                              # targets end with a '`' character
    )
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse parial and complete "default" roles.

A "default" role is the target part of a normal role - but without the ``:name:`` part.
"""


@dataclass
class Role:
    """Represents a role."""

    @dataclass
    class TargetProvider:
        """A target provider instance."""

        name: str
        """The name of the provider."""

        kwargs: dict[str, Any] = field(default_factory=dict)
        """Arguments to pass to the target provider."""

    name: str
    """The name of the role, as the user would type in an rst file."""

    implementation: str | None
    """The dotted name of the role's implementation."""

    location: Location | None = field(default=None)
    """The location of the role's implementation, if known."""

    target_providers: list[TargetProvider] = field(default_factory=list)
    """The list of target providers that can be used with this role."""

    def to_db(
        self, dumps: Callable[[Any], str]
    ) -> tuple[str, str | None, str | None, str | None]:
        """Convert this role to its database representation."""
        if len(self.target_providers) > 0:
            providers = dumps(self.target_providers)
        else:
            providers = None

        location = dumps(self.location) if self.location is not None else None
        return (self.name, self.implementation, location, providers)

    @classmethod
    def from_db(
        cls,
        load_as: JsonLoader,
        name: str,
        implementation: str | None,
        location: str | None,
        providers: str | None,
    ) -> Role:
        """Create a role from its database representation."""

        loc = load_as(location, Location) if location is not None else None
        target_providers = (
            load_as(providers, list[Role.TargetProvider])
            if providers is not None
            else []
        )

        return cls(
            name=name,
            implementation=implementation,
            location=loc,
            target_providers=target_providers,
        )



from __future__ import annotations

import pytest

from esbonio import server
from esbonio.server.features.log import LoggerConfiguration
from esbonio.server.features.log import LoggingConfig

SERVER = server.EsbonioLanguageServer


@pytest.mark.parametrize(
    "config,expected",
    [
        (  # Check the defaults
            LoggingConfig(),
            dict(
                version=1,
                disable_existing_loggers=True,
                formatters=dict(
                    fmt01=dict(format="[%(name)s] %(message)s"),
                    fmt02=dict(format="%(message)s"),
                ),
                handlers=dict(
                    stderr01={
                        "class": "logging.StreamHandler",
                        "level": "DEBUG",
                        "formatter": "fmt01",
                        "stream": "ext://sys.stderr",
                    },
                    stderr02={
                        "class": "logging.StreamHandler",
                        "level": "DEBUG",
                        "formatter": "fmt02",
                        "stream": "ext://sys.stderr",
                    },
                ),
                loggers=dict(
                    esbonio=dict(
                        level="ERROR",
                        propagate=False,
                        handlers=["stderr01"],
                    ),
                    sphinx=dict(
                        level="INFO",
                        propagate=False,
                        handlers=["stderr02"],
                    ),
                ),
            ),
        ),
        (  # It should be possible to override the base logging level
            LoggingConfig(level="debug"),
            dict(
                version=1,
                disable_existing_loggers=True,
                formatters=dict(
                    fmt01=dict(format="[%(name)s] %(message)s"),
                    fmt02=dict(format="%(message)s"),
                ),
                handlers=dict(
                    stderr01={
                        "class": "logging.StreamHandler",
                        "level": "DEBUG",
                        "formatter": "fmt01",
                        "stream": "ext://sys.stderr",
                    },
                    stderr02={
                        "class": "logging.StreamHandler",
                        "level": "DEBUG",
                        "formatter": "fmt02",
                        "stream": "ext://sys.stderr",
                    },
                ),
                loggers=dict(
                    esbonio=dict(
                        level="DEBUG",
                        propagate=False,
                        handlers=["stderr01"],
                    ),
                    sphinx=dict(
                        level="INFO",
                        propagate=False,
                        handlers=["stderr02"],
                    ),
                ),
            ),
        ),
        (  # It should be possible to override the base log format
            LoggingConfig(format="%(message)s"),
            dict(
                version=1,
                disable_existing_loggers=True,
                formatters=dict(
                    fmt01=dict(format="%(message)s"),
                ),
                handlers=dict(
                    stderr01={
                        "class": "logging.StreamHandler",
                        "level": "DEBUG",
                        "formatter": "fmt01",
                        "stream": "ext://sys.stderr",
                    },
                ),
                loggers=dict(
                    esbonio=dict(
                        level="ERROR",
                        propagate=False,
                        handlers=["stderr01"],
                    ),
                    sphinx=dict(
                        level="INFO",
                        propagate=False,
                        handlers=["stderr01"],
                    ),
                ),
            ),
        ),
        (  # User should be able to re-direct output to ``window/logMessages``
            LoggingConfig(stderr=False, window=True),
            dict(
                version=1,
                disable_existing_loggers=True,
                formatters=dict(
                    fmt01=dict(format="[%(name)s] %(message)s"),
                    fmt02=dict(format="%(message)s"),
                ),
                handlers=dict(
                    window01={
                        "()": "esbonio.server.features.log.WindowLogMessageHandler",
                        "level": "DEBUG",
                        "formatter": "fmt01",
                        "server": SERVER,
                    },
                    window02={
                        "()": "esbonio.server.features.log.WindowLogMessageHandler",
                        "level": "DEBUG",
                        "formatter": "fmt02",
                        "server": SERVER,
                    },
                ),
                loggers=dict(
                    esbonio=dict(
                        level="ERROR",
                        propagate=False,
                        handlers=["window01"],
                    ),
                    sphinx=dict(
                        level="INFO",
                        propagate=False,
                        handlers=["window02"],
                    ),
                ),
            ),
        ),
        (  # User should be able to re-direct output to a file
            LoggingConfig(stderr=False, filepath="esbonio.log"),
            dict(
                version=1,
                disable_existing_loggers=True,
                formatters=dict(
                    fmt01=dict(format="[%(name)s] %(message)s"),
                    fmt02=dict(format="%(message)s"),
                ),
                handlers=dict(
                    file01={
                        "class": "logging.FileHandler",
                        "level": "DEBUG",
                        "formatter": "fmt01",
                        "filename": "esbonio.log",
                    },
                    file02={
                        "class": "logging.FileHandler",
                        "level": "DEBUG",
                        "formatter": "fmt02",
                        "filename": "esbonio.log",
                    },
                ),
                loggers=dict(
                    esbonio=dict(
                        level="ERROR",
                        propagate=False,
                        handlers=["file01"],
                    ),
                    sphinx=dict(
                        level="INFO",
                        propagate=False,
                        handlers=["file02"],
                    ),
                ),
            ),
        ),
        (  # User should be able to log to everything if they so desired
            LoggingConfig(stderr=True, window=True, filepath="esbonio.log"),
            dict(
                version=1,
                disable_existing_loggers=True,
                formatters=dict(
                    fmt01=dict(format="[%(name)s] %(message)s"),
                    fmt02=dict(format="%(message)s"),
                ),
                handlers=dict(
                    file01={
                        "class": "logging.FileHandler",
                        "level": "DEBUG",
                        "formatter": "fmt01",
                        "filename": "esbonio.log",
                    },
                    stderr02={
                        "class": "logging.StreamHandler",
                        "level": "DEBUG",
                        "formatter": "fmt01",
                        "stream": "ext://sys.stderr",
                    },
                    window03={
                        "()": "esbonio.server.features.log.WindowLogMessageHandler",
                        "level": "DEBUG",
                        "formatter": "fmt01",
                        "server": SERVER,
                    },
                    file04={
                        "class": "logging.FileHandler",
                        "level": "DEBUG",
                        "formatter": "fmt02",
                        "filename": "esbonio.log",
                    },
                    stderr05={
                        "class": "logging.StreamHandler",
                        "level": "DEBUG",
                        "formatter": "fmt02",
                        "stream": "ext://sys.stderr",
                    },
                    window06={
                        "()": "esbonio.server.features.log.WindowLogMessageHandler",
                        "level": "DEBUG",
                        "formatter": "fmt02",
                        "server": SERVER,
                    },
                ),
                loggers=dict(
                    esbonio=dict(
                        level="ERROR",
                        propagate=False,
                        handlers=["file01", "stderr02", "window03"],
                    ),
                    sphinx=dict(
                        level="INFO",
                        propagate=False,
                        handlers=["file04", "stderr05", "window06"],
                    ),
                ),
            ),
        ),
        (  # The user should be able to customize an individual logger.
            LoggingConfig(
                config={"esbonio.Configuration": LoggerConfiguration(level="info")},
            ),
            dict(
                version=1,
                disable_existing_loggers=True,
                formatters=dict(
                    fmt01=dict(format="[%(name)s] %(message)s"),
                    fmt02=dict(format="%(message)s"),
                ),
                handlers=dict(
                    stderr01={
                        "class": "logging.StreamHandler",
                        "level": "DEBUG",
                        "formatter": "fmt01",
                        "stream": "ext://sys.stderr",
                    },
                    stderr02={
                        "class": "logging.StreamHandler",
                        "level": "DEBUG",
                        "formatter": "fmt02",
                        "stream": "ext://sys.stderr",
                    },
                ),
                loggers={
                    "esbonio": dict(
                        level="ERROR",
                        propagate=False,
                        handlers=["stderr01"],
                    ),
                    "esbonio.Configuration": dict(
                        level="INFO",
                        propagate=False,
                        handlers=["stderr01"],
                    ),
                    "sphinx": dict(
                        level="INFO",
                        propagate=False,
                        handlers=["stderr02"],
                    ),
                },
            ),
        ),
    ],
)
def test_logging_config(config: LoggingConfig, expected: dict):
    """Ensure that we can convert the user's config into the config we can pass to the
    ``logging.config`` module correctly."""

    # SERVER in this case is a class, at runtime this will be an actual instance.
    assert config.to_logging_config(SERVER) == expected



from __future__ import annotations

from lsprotocol import types

from esbonio import server
from esbonio.server.features.directives import DirectiveFeature
from esbonio.server.features.directives import completion
from esbonio.sphinx_agent.types import RST_DIRECTIVE


class RstDirectives(server.LanguageFeature):
    """A frontend to directives for reStructuredText syntax."""

    def __init__(self, directives: DirectiveFeature, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.directives = directives
        self._insert_behavior = "replace"

    completion_trigger = server.CompletionTrigger(
        patterns=[RST_DIRECTIVE],
        languages={"rst"},
        characters={".", "`"},
    )

    definition_trigger = server.DefinitionTrigger(
        patterns=[RST_DIRECTIVE],
        languages={"rst"},
    )

    def initialized(self, params: types.InitializedParams):
        """Called once the initial handshake between client and server has finished."""
        self.configuration.subscribe(
            "esbonio.server.completion",
            server.CompletionConfig,
            self.update_configuration,
        )

    def update_configuration(
        self, event: server.ConfigChangeEvent[server.CompletionConfig]
    ):
        """Called when the user's configuration is updated."""
        self._insert_behavior = event.value.preferred_insert_behavior

    async def completion(
        self, context: server.CompletionContext
    ) -> list[types.CompletionItem] | None:
        """Provide completion suggestions for directives."""

        groups = context.match.groupdict()

        # Are we completing a directive's options?
        if "directive" not in groups:
            return await self.complete_options(context)

        # Don't offer completions for targets
        if (groups["name"] or "").startswith("_"):
            return None

        # Are we completing the directive's argument?
        directive_end = context.match.span()[0] + len(groups["directive"])
        complete_directive = groups["directive"].endswith("::")

        if complete_directive and directive_end < context.position.character:
            return await self.complete_arguments(context)

        return await self.complete_directives(context)

    async def complete_options(self, context: server.CompletionContext):
        return None

    async def complete_arguments(
        self, context: server.CompletionContext
    ) -> list[types.CompletionItem] | None:
        """Return completion suggestions for the current directive's argument"""

        render_func = completion.get_directive_argument_renderer(
            context.language, self._insert_behavior
        )
        if render_func is None:
            return None

        items = []
        directive_name = context.match.group("name")
        suggestions = await self.directives.suggest_arguments(context, directive_name)

        for argument in suggestions:
            if (item := render_func(context, argument)) is not None:
                items.append(item)

        return items if len(items) > 0 else None

    async def complete_directives(
        self, context: server.CompletionContext
    ) -> list[types.CompletionItem] | None:
        """Return completion suggestions for the available directives."""

        render_func = completion.get_directive_renderer(
            context.language, self._insert_behavior
        )
        if render_func is None:
            return None

        items = []
        for directive in await self.directives.suggest_directives(context):
            if (item := render_func(context, directive)) is not None:
                items.append(item)

        if len(items) > 0:
            return items

        return None

    async def definition(
        self, context: server.DefinitionContext
    ) -> list[types.Location] | None:
        """Find the definition of the requested item"""
        directive = context.match.group("name")
        argument = context.match.group("argument")

        if not argument:
            return None

        start = context.match.group(0).index(argument)
        end = start + len(argument)

        if start <= context.position.character <= end:
            return await self.directives.find_argument_definition(
                context, directive, argument
            )

        return None

    async def document_link(
        self, context: server.DocumentLinkContext
    ) -> list[types.DocumentLink] | None:
        links = []

        for line, text in enumerate(context.doc.lines):
            for match in RST_DIRECTIVE.finditer(text):
                if "::" not in match.group("directive"):
                    continue

                if not (argument := match.group("argument")):
                    continue

                name = match.group("name")
                target = await self.directives.resolve_argument_link(
                    context, name, argument
                )

                if target is None:
                    continue

                tooltip = None
                if isinstance(target, tuple):
                    target, tooltip = target

                idx = match.group(0).index(argument)
                start = match.start() + idx
                end = start + len(argument)

                links.append(
                    types.DocumentLink(
                        target=target,
                        tooltip=tooltip if context.tooltip_support else None,
                        range=types.Range(
                            start=types.Position(line=line, character=start),
                            end=types.Position(line=line, character=end),
                        ),
                    )
                )

        return links if len(links) > 0 else None


def esbonio_setup(esbonio: server.EsbonioLanguageServer, directives: DirectiveFeature):
    rst_directives = RstDirectives(directives, esbonio)
    esbonio.add_feature(rst_directives)
